from io import BytesIO
from PIL import Image
from typing import Any
from typing import Dict
from typing import Union
from pydantic import BaseModel
from cfclient.core import HttpClient
from cfclient.core import ClientSession


def get_url(host: str, endpoint: str) -> str:
    return f"{host}{endpoint}"


class get_http_session:
    def __init__(self):
        self._http_client = None

    async def __aenter__(self) -> ClientSession:
        self._http_client = HttpClient()
        self._http_client.start()
        return self._http_client.session

    async def __aexit__(self, *args: Any) -> None:
        await self._http_client.stop()
        self._http_client = None


async def get_image_res(url: str, d: Union[BaseModel, Dict[str, Any]]) -> Image.Image:
    if isinstance(d, BaseModel):
        d = d.dict()
    async with get_http_session() as session:
        async with session.post(url, json=d) as response:
            return Image.open(BytesIO(await response.read()))


__all__ = [
    "get_url",
    "get_http_session",
    "get_image_res",
]



import time

from typing import Any
from typing import List
from pydantic import Field
from pydantic import BaseModel
from cflearn.api.cv.third_party.prompt import PromptEnhanceConfig

from .utils import api_pool
from .utils import APIs
from .common import register_prompt_enhance
from .common import IAlgorithm
from .common import TextModel


txt2txt_prompt_enhance_endpoint = "/txt2txt/prompt_enhance"


class PromptEnhanceModel(TextModel):
    temperature: float = Field(
        0.9,
        description="a higher temperature will produce more diverse results, but with a higher risk of less coherent text.",
    )
    top_k: int = Field(
        8,
        description="the number of tokens to sample from at each step.",
    )
    max_length: int = Field(
        76,
        description="the maximum number of tokens for the output of the model.",
    )
    repitition_penalty: float = Field(
        1.2,
        description="the penalty value for each repetition of a token.",
    )
    num_return_sequences: int = Field(
        1,
        description="the number of results to generate.",
    )
    comma_mode: bool = Field(False, description="whether include more comma.")


class PromptEnhanceResponse(BaseModel):
    prompts: List[str] = Field(..., description="enhanced prompts.")


@IAlgorithm.auto_register()
class Txt2TxtPromptEnhance(IAlgorithm):
    model_class = TextModel
    response_model_class = PromptEnhanceResponse

    endpoint = txt2txt_prompt_enhance_endpoint

    def initialize(self) -> None:
        register_prompt_enhance()

    async def run(self, data: PromptEnhanceModel, *args: Any) -> PromptEnhanceResponse:
        self.log_endpoint(data)
        t0 = time.time()
        m = api_pool.get(APIs.PROMPT_ENHANCE)
        t1 = time.time()
        kw = data.dict()
        text = kw.pop("text")
        prompts = m.enhance(text, config=PromptEnhanceConfig(**kw))
        t2 = time.time()
        api_pool.cleanup(APIs.PROMPT_ENHANCE)
        self.log_times(
            {
                "get_model": t1 - t0,
                "inference": t2 - t1,
                "cleanup": time.time() - t2,
            }
        )
        return PromptEnhanceResponse(prompts=prompts)


__all__ = [
    "txt2txt_prompt_enhance_endpoint",
    "Txt2TxtPromptEnhance",
]



import cv2
import math
import torch

import numpy as np

from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Protocol
from cflearn.api.utils import ILoadableItem
from cflearn.api.utils import ILoadablePool

from .parameters import lazy_load
from .parameters import pool_limit
from .parameters import OPT


def resize_image(input_image: np.ndarray, resolution: int) -> np.ndarray:
    H, W, C = input_image.shape
    H = float(H)
    W = float(W)
    k = float(resolution) / min(H, W)
    H *= k
    W *= k
    H = int(np.round(H / 64.0)) * 64
    W = int(np.round(W / 64.0)) * 64
    img = cv2.resize(
        input_image,
        (W, H),
        interpolation=cv2.INTER_LANCZOS4 if k > 1 else cv2.INTER_AREA,
    )
    return img


def to_canvas(results: List[np.ndarray], *, padding: int = 0) -> np.ndarray:
    num_results = len(results)
    if num_results == 1:
        return results[0]
    num_col = math.ceil(math.sqrt(num_results))
    num_row = round(num_results / num_col)
    if num_row * num_col < num_results:
        num_row += 1
    h, w = results[0].shape[:2]
    canvas_w = num_col * w + (num_col - 1) * padding
    canvas_h = num_row * h + (num_row - 1) * padding
    canvas = np.full([canvas_h, canvas_w, 3], 255, np.uint8)
    for i, out in enumerate(results):
        ih, iw = out.shape[:2]
        if h != ih:
            raise ValueError(f"`h` mismatch: {ih} != {h}")
        if w != iw:
            raise ValueError(f"`w` mismatchh: {iw} != {w}")
        ix = i % num_col
        iy = i // num_col
        ix = ix * w + ix * padding
        iy = iy * h + iy * padding
        canvas[iy : iy + h, ix : ix + w] = out
    return canvas


class APIs(str, Enum):
    SD = "sd"
    SD_INPAINTING = "sd_inpainting"
    ESR = "esr"
    ESR_ANIME = "esr_anime"
    INPAINTING = "inpainting"
    LAMA = "lama"
    SEMANTIC = "semantic"
    HRNET = "hrnet"
    ISNET = "isnet"
    BLIP = "blip"
    PROMPT_ENHANCE = "prompt_enhance"


class IAPI:
    def to(self, device: str, *, use_half: bool) -> None:
        pass


class APIInit(Protocol):
    def __call__(self, init_to_cpu: bool) -> IAPI:
        pass


class LoadableAPI(ILoadableItem[IAPI]):
    def __init__(
        self,
        init_fn: APIInit,
        *,
        init: bool = False,
        is_sd: bool = False,
    ):
        super().__init__(lambda: init_fn(self.init_to_cpu), init=init)
        self.is_sd = is_sd

    @property
    def lazy(self) -> bool:
        return lazy_load() and not self.is_sd

    @property
    def init_to_cpu(self) -> bool:
        return self.lazy or OPT["cpu"]

    @property
    def need_change_device(self) -> bool:
        return self.lazy and not OPT["cpu"]

    @property
    def sd_kwargs(self) -> Dict[str, Any]:
        return {"no_annotator": True} if self.is_sd else {}

    def load(self, *, no_change: bool = False, **kwargs: Any) -> IAPI:
        super().load()
        if not no_change and self.need_change_device:
            self._item.to("cuda:0", use_half=True, **self.sd_kwargs)
        return self._item

    def cleanup(self) -> None:
        if self.need_change_device:
            self._item.to("cpu", use_half=False, **self.sd_kwargs)
            torch.cuda.empty_cache()

    def unload(self) -> None:
        self.cleanup()
        return super().unload()


class APIPool(ILoadablePool[IAPI]):
    def register(self, key: str, init_fn: APIInit) -> None:
        def _init(init: bool) -> LoadableAPI:
            is_sd = key in (APIs.SD, APIs.SD_INPAINTING)
            api = LoadableAPI(init_fn, init=False, is_sd=is_sd)
            print("> init", key, "(lazy)" if api.lazy else "")
            if init:
                api.load(no_change=api.lazy)
            return api

        if key in self:
            return
        return super().register(key, _init)

    def cleanup(self, key: str) -> None:
        loadable_api: Optional[LoadableAPI] = self.pool.get(key)
        if loadable_api is None:
            raise ValueError(f"key '{key}' does not exist")
        loadable_api.cleanup()

    def need_change_device(self, key: str) -> bool:
        loadable_api: Optional[LoadableAPI] = self.pool.get(key)
        if loadable_api is None:
            raise ValueError(f"key '{key}' does not exist")
        return loadable_api.need_change_device

    def update_limit(self) -> None:
        self.limit = pool_limit()


api_pool = APIPool()



import os
import json
import time
import yaml
import redis
import torch
import asyncio
import datetime
import logging.config

import numpy as np

from kafka import KafkaConsumer
from typing import Any
from typing import Dict
from typing import Union
from fastapi import Response
from qcloud_cos import CosConfig
from qcloud_cos import CosS3Client

from cfclient.models import *
from cftool.misc import shallow_copy_dict
from cfclient.core import HttpClient
from cfclient.core import TritonClient
from cfclient.utils import get_err_msg
from cfclient.utils import run_algorithm

# This is necessary to register the algorithms
from cfcreator import *


# logging
root = os.path.dirname(__file__)
logging_root = os.path.join(root, "logs", "consumer")
os.makedirs(logging_root, exist_ok=True)
with open(os.path.join(root, "config.yml")) as f:
    now = datetime.datetime.now()
    timestamp = now.strftime("%Y-%m-%d_%H-%M-%S-%f")
    log_path = os.path.join(logging_root, f"{timestamp}.log")
    config = yaml.load(f, Loader=yaml.FullLoader)
    config["handlers"]["file"]["filename"] = log_path
    logging.config.dictConfig(config)

logging.getLogger("aiokafka").disabled = True
logging.getLogger("kafka.conn").disabled = True
logging.getLogger("kafka.client").disabled = True
logging.getLogger("kafka.cluster").disabled = True
logging.getLogger("kafka.coordinator").disabled = True
logging.getLogger("kafka.coordinator.consumer").disabled = True
logging.getLogger("kafka.metrics.metrics").disabled = True
logging.getLogger("kafka.protocol.parser").disabled = True
logging.getLogger("kafka.consumer.fetcher").disabled = True
logging.getLogger("kafka.consumer.group_coordinator").disabled = True
logging.getLogger("kafka.consumer.subscription_state").disabled = True


constants = dict(
    triton_host=None,
    triton_port=8000,
)


# clients
## cos client
config = CosConfig(
    Region=REGION,
    SecretId=SECRET_ID,
    SecretKey=SECRET_KEY,
    Scheme=SCHEME,
)
cos_client = CosS3Client(config)
## http client
http_client = HttpClient()
## triton client
triton_host = constants["triton_host"]
if triton_host is None:
    triton_client = None
else:
    triton_client = TritonClient(url=f"{triton_host}:{constants['triton_port']}")
## collect
clients = dict(
    http=http_client,
    triton=triton_client,
)

redis_client = redis.Redis(**redis_kwargs())
audit_redis_client = redis.Redis(**audit_redis_kwargs())
pending_queue_key = get_pending_queue_key()


# algorithms
loaded_algorithms: Dict[str, IAlgorithm] = {
    k: v(clients) for k, v in algorithms.items()
}


def get_redis_number(key: str) -> int:
    data = redis_client.get(key)
    if data is None:
        return 0
    return int(data.decode())  # type: ignore


def get_pending_queue() -> list:
    data = redis_client.get(pending_queue_key)
    if data is None:
        return []
    return json.loads(data)


async def post_callback(
    url: str,
    uid: str,
    success: bool,
    data: Dict[str, Any],
) -> None:
    if url:
        try:
            data = dict(uid=uid, success=success, data=data)
            async with http_client.session.post(url, json=data, timeout=1) as res:
                await res.json()
        except Exception as err:
            print(
                f"\n\n!!! post to callback_url ({url}) failed "
                f"({get_err_msg(err)}) !!!\n\n"
            )


def simplify(params: Dict[str, Any]) -> Dict[str, Any]:
    params = shallow_copy_dict(params)
    custom_embeddings = params.get("custom_embeddings")
    if custom_embeddings is not None and isinstance(custom_embeddings, dict):
        simplified_embeddings = {}
        for k, v in custom_embeddings.items():
            try:
                v_array = np.atleast_2d(v)
                if v_array.shape[1] <= 6:
                    simplified_embeddings[k] = v
                else:
                    v_array = np.concatenate([v_array[:, :3], v_array[:, -3:]], axis=1)
                    simplified_embeddings[k] = v_array.tolist()
            except Exception as err:
                simplified_embeddings[k] = get_err_msg(err)
        params["custom_embeddings"] = simplified_embeddings
    return params


# kafka & redis
async def consume() -> None:
    OPT["verbose"] = False
    OPT["lazy_load"] = True

    topic = kafka_topic()
    expire_seconds = 10 * 365 * 24 * 3600

    redis_client.expire(pending_queue_key, expire_seconds)
    # initialize
    http_client.start()
    for v in loaded_algorithms.values():
        v.initialize()
    print("> Algorithms are Loaded!")
    kafka_consumer = KafkaConsumer(
        topic,
        group_id=kafka_group_id(),
        bootstrap_servers=kafka_server(),
        max_poll_records=kafka_max_poll_records(),
        max_poll_interval_ms=kafka_max_poll_interval_ms(),
    )
    # main loop
    try:
        for message in kafka_consumer:
            data = json.loads(message.value)
            uid = data["uid"]
            task = data["task"]
            if task == "scene-generation":
                task = "txt2img.sd.outpainting"
            params = data["params"]
            callback_url = params.get("callback_url", "")
            existing = redis_client.get(uid)
            if existing is not None:
                existing = json.loads(existing)
                print(">>> existing", existing)
                if existing["status"] in (
                    Status.FINISHED,
                    Status.EXCEPTION,
                    Status.INTERRUPTED,
                ):
                    continue
            print(">>> working", uid)
            data = {} if existing is None else (existing.get("data", {}) or {})
            start_time = time.time()
            data["start_time"] = start_time
            create_time = data.get("create_time", start_time)
            redis_client.set(uid, json.dumps(dict(status=Status.WORKING, data=data)))
            procedure = "start"
            model = None
            try:
                algorithm = loaded_algorithms[task]
                model = algorithm.model_class(**params)  # type: ignore
                procedure = "start -> run_algorithm"
                if isinstance(model, ReturnArraysModel):
                    model.return_arrays = True
                res: Union[Response, Any] = await run_algorithm(algorithm, model)
                latencies = algorithm.last_latencies
                t1 = time.time()
                if task.startswith("control") or task.startswith("pipeline"):
                    procedure = "run_algorithm -> upload_temp_image"
                    url_results = [upload_temp_image(cos_client, arr) for arr in res]
                    urls = [rs.cdn for rs in url_results]
                    t2 = time.time()
                    procedure = "upload_temp_image -> audit_image"
                    if not isinstance(model, ControlNetModel) or not model.use_audit:
                        reasons = [""] * len(urls)
                    else:
                        reasons = []
                        for i, rs in enumerate(url_results):
                            try:
                                audit = audit_image(
                                    cos_client, image_mod_client, rs.path
                                )
                            except:
                                audit = AuditResponse(safe=False, reason="unknown")
                            if audit.safe:
                                reasons.append("")
                            else:
                                urls[i] = None
                                reasons.append(audit.reason)
                    t3 = time.time()
                    procedure = "audit_image -> redis"
                    if task.startswith("control"):
                        types = params.get("types")
                        num_cond = 1 if types is None else len(types)
                        result = dict(
                            uid=uid,
                            response=dict(
                                hint_urls=urls[:num_cond],
                                hint_reasons=reasons[:num_cond],
                                result_urls=urls[num_cond:],
                                result_reasons=reasons[num_cond:],
                            ),
                        )
                    elif task == "pipeline.paste":
                        result = dict(uid=uid, response=dict(url=urls[0]))
                    else:
                        raise ValueError(f"unrecognized task '{task}' occurred")
                elif algorithm.response_model_class is not None:
                    t2 = t3 = t1
                    procedure = "run_algorithm -> redis"
                    result = dict(uid=uid, response=res.dict())
                else:
                    procedure = "run_algorithm -> upload_temp_image"
                    urls = upload_temp_image(cos_client, res.body)
                    t2 = time.time()
                    procedure = "upload_temp_image -> audit_image"
                    if task != "img2img.sr":
                        try:
                            audit = audit_image(audit_redis_client, urls.path)
                        except:
                            audit = AuditResponse(safe=False, reason="unknown")
                    else:
                        audit = AuditResponse(safe=True, reason="")
                    t3 = time.time()
                    procedure = "audit_image -> redis"
                    result = dict(
                        uid=uid,
                        cdn=urls.cdn if audit.safe else "",
                        cos=urls.cos if audit.safe else "",
                        safe=audit.safe,
                        reason=audit.reason,
                    )
                result.update(data)
                end_time = time.time()
                result["end_time"] = end_time
                result["duration"] = end_time - create_time
                result["elapsed_times"] = dict(
                    pending=start_time - create_time,
                    run_algorithm=t1 - start_time,
                    algorithm_latencies=latencies,
                    upload=t2 - t1,
                    audit=t3 - t2,
                )
                result["request"] = dict(task=task, model=simplify(model.dict()))
                redis_client.set(
                    uid,
                    json.dumps(dict(status=Status.FINISHED, data=result)),
                )
                t4 = time.time()
                procedure = "redis -> callback"
                await post_callback(callback_url, uid, True, result)
                t5 = time.time()
                procedure = "callback -> update_elapsed_times"
                result["elapsed_times"].update(
                    dict(
                        redis=t4 - t3,
                        callback=t5 - t4,
                    )
                )
                redis_client.set(
                    uid,
                    json.dumps(dict(status=Status.FINISHED, data=result)),
                )
                procedure = "done"
                # maintain queue
                queue = get_pending_queue()
                if uid in queue:
                    queue.remove(uid)
                    redis_client.set(pending_queue_key, json.dumps(queue))
            except Exception as err:
                end_time = time.time()
                torch.cuda.empty_cache()
                reason = f"{task} -> {procedure} : {get_err_msg(err)}"
                data["uid"] = uid
                data["reason"] = reason
                data["end_time"] = end_time
                data["duration"] = end_time - create_time
                if model is None:
                    data["request"] = dict(task=task, params=simplify(params))
                else:
                    data["request"] = dict(task=task, model=simplify(model.dict()))
                redis_client.set(
                    uid,
                    json.dumps(
                        dict(status=Status.EXCEPTION, data=data),
                        ensure_ascii=False,
                    ),
                )
                await post_callback(callback_url, uid, False, data)
    finally:
        # clean up
        await http_client.stop()
        print(">>> end")


asyncio.run(consume())



import os
import click
import uvicorn

from cfcreator import opt_env_context
from cfcreator import Focus
from cflearn.parameters import OPT


@click.group()
def main() -> None:
    pass


@main.command()
@click.option(
    "-p",
    "--port",
    default=8123,
    show_default=True,
    type=int,
    help="Port of the service.",
)
@click.option(
    "--cpu",
    is_flag=True,
    default=False,
    show_default=True,
    type=bool,
    help="This flag will force the server to use cpu only.",
)
@click.option(
    "--focus",
    default="all",
    show_default=True,
    type=click.Choice([e.value for e in Focus]),
    help="""
Indicates which endpoints should we focus on, helpful if we only care about certain subset of features.
\n- all            |  will load all endpoints.
\n- sd             |  will load SD endpoints.
\n- sd.base        |  will only load basic SD endpoints, which means anime / inpainting / outpainting endpoints will not be loaded.
\n- sd.anime       |  will only load anime SD endpoints, which means basic / inpainting / outpainting endpoints will not be loaded.
\n- sd.inpainting  |  will only load inpainting / outpainting SD endpoints, which means basic / anime endpoints will not be loaded.
\n- sync           |  will only load 'sync' endpoints, which are relatively fast (e.g. sod, lama, captioning, harmonization, ...).
\n- control        |  will only load 'ControlNet' endpoints.
\n- pipeline       |  will only load 'Pipeline' endpoints.
\n-
""",
)
@click.option(
    "-r",
    "--reload",
    is_flag=True,
    default=False,
    show_default=True,
    type=bool,
    help="This flag represents the `reload` argument of `uvicorn.run`.",
)
@click.option(
    "-d",
    "--cache_dir",
    default="",
    show_default=True,
    type=str,
    help="Directory of the cache files.",
)
@click.option(
    "--lazy",
    is_flag=True,
    default=False,
    show_default=True,
    type=bool,
    help="This flag will enable the `lazy loading` mode.",
)
@click.option(
    "--limit",
    default=-1,
    show_default=True,
    type=int,
    help="Limitation of the pools (i.e. `WeightsPool` & `APIPool`).",
)
def serve(
    *,
    port: int,
    cpu: bool,
    focus: str,
    reload: bool,
    cache_dir: str,
    lazy: bool,
    limit: int,
) -> None:
    increment = dict(lazy_load=lazy, pool_limit=limit)
    if cpu:
        increment["cpu"] = True
    if focus != "all":
        increment["focus"] = focus
    cflearn_increment = {}
    if cache_dir:
        cflearn_increment["cache_dir"] = cache_dir
    with opt_env_context(increment):
        with OPT.opt_context(cflearn_increment):
            uvicorn.run(
                "cfcreator.apis.interface:app",
                host="0.0.0.0",
                port=port,
                reload=reload,
                reload_dirs=os.path.dirname(__file__) if reload else None,
            )



import time

from PIL import Image
from typing import Any
from cftool.cv import to_rgb
from cftool.cv import restrict_wh

from .utils import api_pool
from .utils import APIs
from .common import register_blip
from .common import IAlgorithm
from .common import TextModel
from .common import ImageModel
from .common import MaxWHModel


img2txt_caption_endpoint = "/img2txt/caption"


class Img2TxtModel(MaxWHModel, ImageModel):
    pass


@IAlgorithm.auto_register()
class Img2TxtCaption(IAlgorithm):
    model_class = Img2TxtModel
    response_model_class = TextModel

    endpoint = img2txt_caption_endpoint

    def initialize(self) -> None:
        register_blip()

    async def run(self, data: Img2TxtModel, *args: Any) -> TextModel:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.download_image_with_retry(data.url)
        t1 = time.time()
        w, h = image.size
        w, h = restrict_wh(w, h, data.max_wh)
        image = image.resize((w, h), resample=Image.LANCZOS)
        t2 = time.time()
        m = api_pool.get(APIs.BLIP)
        t3 = time.time()
        caption = m.caption(to_rgb(image))
        t4 = time.time()
        api_pool.cleanup(APIs.BLIP)
        self.log_times(
            {
                "download": t1 - t0,
                "preprocess": t2 - t1,
                "get_model": t3 - t2,
                "inference": t4 - t3,
                "cleanup": time.time() - t4,
            }
        )
        return TextModel(text=caption)


__all__ = [
    "img2txt_caption_endpoint",
    "Img2TxtCaption",
]



import time

from enum import Enum
from typing import Any
from fastapi import Response
from pydantic import Field
from pydantic import BaseModel
from cfclient.models import ImageModel

from .utils import api_pool
from .utils import APIs
from .common import register_sd
from .common import register_sd_inpainting
from .common import get_sd_from
from .common import handle_diffusion_model
from .common import get_bytes_from_diffusion
from .common import IAlgorithm
from .common import Txt2ImgModel
from .common import CommonSDInpaintingModel


txt2img_sd_endpoint = "/txt2img/sd"
txt2img_sd_inpainting_endpoint = "/txt2img/sd.inpainting"
txt2img_sd_outpainting_endpoint = "/txt2img/sd.outpainting"


class _Txt2ImgSDModel(BaseModel):
    w: int = Field(512, description="The desired output width.")
    h: int = Field(512, description="The desired output height.")


class Txt2ImgSDModel(Txt2ImgModel, _Txt2ImgSDModel):
    pass


@IAlgorithm.auto_register()
class Txt2ImgSD(IAlgorithm):
    model_class = Txt2ImgSDModel

    endpoint = txt2img_sd_endpoint

    def initialize(self) -> None:
        register_sd()

    async def run(self, data: Txt2ImgSDModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        m = get_sd_from(data)
        t1 = time.time()
        size = data.w, data.h
        kwargs = handle_diffusion_model(m, data)
        img_arr = m.txt2img(
            data.text,
            size=size,
            max_wh=data.max_wh,
            **kwargs,
        ).numpy()[0]
        content = get_bytes_from_diffusion(img_arr)
        t2 = time.time()
        api_pool.cleanup(APIs.SD)
        self.log_times(
            {
                "get_model": t1 - t0,
                "inference": t2 - t1,
                "cleanup": time.time() - t2,
            }
        )
        return Response(content=content, media_type="image/png")


class PaddingModes(str, Enum):
    CV2_NS = "cv2_ns"
    CV2_TELEA = "cv2_telea"


class Txt2ImgSDInpaintingModel(CommonSDInpaintingModel, Txt2ImgModel, ImageModel):
    mask_url: str = Field(
        ...,
        description="""
The `cdn` / `cos` url of the user's mask.
> `cos` url from `qcloud` is preferred.
> If empty string is provided, then we will use an empty mask, which means we will simply perform an image-to-image transform.  
""",
    )


class Txt2ImgSDOutpaintingModel(CommonSDInpaintingModel, Txt2ImgModel, ImageModel):
    pass


@IAlgorithm.auto_register()
class Txt2ImgSDInpainting(IAlgorithm):
    model_class = Txt2ImgSDInpaintingModel

    endpoint = txt2img_sd_inpainting_endpoint

    def initialize(self) -> None:
        register_sd()
        register_sd_inpainting()

    async def run(self, data: Txt2ImgSDInpaintingModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.download_image_with_retry(data.url)
        mask = await self.download_image_with_retry(data.mask_url)
        t1 = time.time()
        if data.use_raw_inpainting:
            api_key = APIs.SD
            m = get_sd_from(data)
        else:
            api_key = APIs.SD_INPAINTING
            m = api_pool.get(api_key)
            m.disable_control()
        t2 = time.time()
        kwargs = handle_diffusion_model(m, data)
        kwargs.update(await self.handle_diffusion_inpainting_model(data))
        img_arr = m.txt2img_inpainting(
            data.text,
            image,
            mask,
            anchor=64,
            max_wh=data.max_wh,
            keep_original=data.keep_original,
            **kwargs,
        ).numpy()[0]
        content = get_bytes_from_diffusion(img_arr)
        t3 = time.time()
        api_pool.cleanup(api_key)
        self.log_times(
            {
                "download": t1 - t0,
                "get_model": t2 - t1,
                "inference": t3 - t2,
                "cleanup": time.time() - t3,
            }
        )
        return Response(content=content, media_type="image/png")


@IAlgorithm.auto_register()
class Txt2ImgSDOutpainting(IAlgorithm):
    model_class = Txt2ImgSDOutpaintingModel

    endpoint = txt2img_sd_outpainting_endpoint

    def initialize(self) -> None:
        register_sd_inpainting()

    async def run(self, data: Txt2ImgSDOutpaintingModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.download_image_with_retry(data.url)
        t1 = time.time()
        m = api_pool.get(APIs.SD_INPAINTING)
        m.disable_control()
        t2 = time.time()
        kwargs = handle_diffusion_model(m, data)
        kwargs.update(await self.handle_diffusion_inpainting_model(data))
        img_arr = m.outpainting(
            data.text,
            image,
            anchor=64,
            max_wh=data.max_wh,
            keep_original=data.keep_original,
            **kwargs,
        ).numpy()[0]
        content = get_bytes_from_diffusion(img_arr)
        t3 = time.time()
        api_pool.cleanup(APIs.SD_INPAINTING)
        self.log_times(
            {
                "download": t1 - t0,
                "get_model": t2 - t1,
                "inference": t3 - t2,
                "cleanup": time.time() - t3,
            }
        )
        return Response(content=content, media_type="image/png")


__all__ = [
    "txt2img_sd_endpoint",
    "txt2img_sd_inpainting_endpoint",
    "txt2img_sd_outpainting_endpoint",
    "Txt2ImgSDModel",
    "Txt2ImgSDInpaintingModel",
    "Txt2ImgSDOutpaintingModel",
    "Txt2ImgSD",
    "Txt2ImgSDInpainting",
    "Txt2ImgSDOutpainting",
]



# `direct` sdk is used to access to the server launched by `cfcreator serve`

from cfcreator import *
from typing import Any
from typing import Dict
from PIL.Image import Image

from .utils import *


class DirectSDK:
    def __init__(self, host: str = "http://localhost:8123") -> None:
        self.host = host

    async def txt2img(self, model: Txt2ImgSDModel) -> Image:
        return await self.push_direct(txt2img_sd_endpoint, model.dict())

    async def push_direct(self, endpoint: str, params: Dict[str, Any]) -> Image:
        return await get_image_res(get_url(self.host, endpoint), params)


__all__ = [
    "DirectSDK",
]



import time

from typing import Any
from typing import Dict
from typing import List
from fastapi import Response
from pydantic import Field
from cftool.cv import np_to_bytes
from cftool.misc import shallow_copy_dict
from cflearn.api.cv.diffusion import ControlNetHints

from .utils import to_canvas
from .utils import APIs
from .common import register_sd
from .common import IAlgorithm
from .common import ControlNetModel
from .control import get_images
from .control import apply_control
from .control import MLSDModel
from .control import PoseModel
from .control import CannyModel
from .control import DepthModel
from .control import ControlMLSDModel
from .control import ControlPoseModel
from .control import ControlCannyModel
from .control import ControlDepthModel


control_multi_endpoint = "/control/multi"
model_mapping = {
    ControlNetHints.MLSD: ControlMLSDModel,
    ControlNetHints.POSE: ControlPoseModel,
    ControlNetHints.CANNY: ControlCannyModel,
    ControlNetHints.DEPTH: ControlDepthModel,
}


class ControlMultiModel(ControlNetModel):
    types: List[ControlNetHints] = Field(..., description="Target control types")
    depth_params: DepthModel = Field(
        DepthModel(),
        description="Params for depth control.",
    )
    canny_params: CannyModel = Field(
        CannyModel(),
        description="Params for canny control.",
    )
    pose_params: PoseModel = Field(
        PoseModel(),
        description="Params for pose control.",
    )
    mlsd_params: MLSDModel = Field(
        MLSDModel(),
        description="Params for mlsd control.",
    )


def gather_all_data(data: ControlMultiModel) -> Dict[ControlNetHints, ControlNetModel]:
    common = data.dict()
    for k in [
        "types",
        "depth_params",
        "canny_params",
        "pose_params",
        "mlsd_params",
    ]:
        common.pop(k)
    all_data = {}
    for hint_type in data.types:
        h_kw = shallow_copy_dict(common)
        h_kw.update(getattr(data, f"{hint_type}_params").dict())
        all_data[hint_type] = model_mapping[hint_type](**h_kw)
    return all_data


@IAlgorithm.auto_register()
class ControlMulti(IAlgorithm):
    model_class = ControlMultiModel

    endpoint = control_multi_endpoint

    def initialize(self) -> None:
        register_sd()

    async def run(self, data: ControlMultiModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image, hint_image = await get_images(self, data)
        t1 = time.time()
        results, latencies = apply_control(
            gather_all_data(data),
            APIs.SD,
            image,
            hint_image,
            data.types,
        )
        t2 = time.time()
        content = None if data.return_arrays else np_to_bytes(to_canvas(results))
        t3 = time.time()
        latencies["download"] = t1 - t0
        latencies["to_canvas"] = t3 - t2
        self.log_times(latencies)
        if data.return_arrays:
            return results
        return Response(content=content, media_type="image/png")


__all__ = [
    "control_multi_endpoint",
    "ControlMultiModel",
    "ControlMulti",
]



import os
import cv2
import time
import torch

import numpy as np

from PIL import Image
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import Optional
from fastapi import Response
from pydantic import Field
from pydantic import BaseModel
from cftool.cv import to_rgb
from cftool.cv import to_uint8
from cftool.cv import np_to_bytes
from cftool.cv import restrict_wh
from cftool.cv import get_suitable_size
from cfclient.models.core import ImageModel
from cflearn.api.cv.diffusion import ControlNetHints
from cflearn.api.cv.diffusion import ControlledDiffusionAPI

from .utils import api_pool
from .utils import to_canvas
from .utils import resize_image
from .utils import APIs
from .common import register_sd
from .common import handle_diffusion_model
from .common import IAlgorithm
from .common import ControlNetModel
from .common import ReturnArraysModel
from .common import ControlStrengthModel


root = os.path.dirname(__file__)
control_depth_endpoint = "/control/depth"
control_canny_endpoint = "/control/canny"
control_pose_endpoint = "/control/pose"
control_mlsd_endpoint = "/control/mlsd"
control_depth_hint_endpoint = "/control/depth_hint"
control_canny_hint_endpoint = "/control/canny_hint"
control_pose_hint_endpoint = "/control/pose_hint"
control_mlsd_hint_endpoint = "/control/mlsd_hint"


images_type = Tuple[np.ndarray, np.ndarray]
apply_response = Tuple[List[np.ndarray], Dict[str, float]]


class ControlNetModelPlaceholder(ControlStrengthModel, ControlNetModel):
    pass


async def get_images(self: IAlgorithm, data: ControlNetModel) -> images_type:
    image = np.array(to_rgb(await self.download_image_with_retry(data.url)))
    if not data.hint_url:
        return image, image
    hint_image = np.array(to_rgb(await self.download_image_with_retry(data.hint_url)))
    return image, hint_image


def apply_control(
    data: Union[ControlNetModelPlaceholder, Dict[str, ControlNetModelPlaceholder]],
    api_key: APIs,
    input_image: np.ndarray,
    hint_image: np.ndarray,
    hint_types: Union[ControlNetHints, List[ControlNetHints]],
    normalized_inpainting_mask: Optional[np.ndarray] = None,
) -> apply_response:
    api: ControlledDiffusionAPI = api_pool.get(api_key, no_change=True)
    need_change_device = api_pool.need_change_device(api_key)
    api.enable_control()
    if not isinstance(data, dict):
        common_data = data
        detect_resolution = getattr(data, "detect_resolution", None)
    else:
        common_data = list(data.values())[0]
        detect_resolution = {}
        for hint_type, h_data in data.items():
            detect_resolution[hint_type] = getattr(h_data, "detect_resolution", None)
    original_h, original_w = input_image.shape[:2]
    resize_to_original = lambda array: cv2.resize(
        array,
        (original_w, original_h),
        interpolation=cv2.INTER_CUBIC,
    )
    w, h = restrict_wh(original_w, original_h, common_data.max_wh)
    w = get_suitable_size(w, 64)
    h = get_suitable_size(h, 64)
    t0 = time.time()
    if not isinstance(hint_types, list):
        hint_types = [hint_types]
    if len(hint_types) > api.num_pool:
        msg = f"maximum number of control is {api.num_pool}, but got {len(hint_types)}"
        raise ValueError(msg)
    api.switch_sd(common_data.base_model)
    api.switch_control(*hint_types)
    t1 = time.time()
    all_hint = {}
    all_o_hint_arrays = []
    all_annotator_change_device_times = []
    for hint_type in sorted(hint_types):
        if detect_resolution is None or isinstance(detect_resolution, int):
            h_res = detect_resolution
        else:
            h_res = detect_resolution.get(hint_type)
        if not isinstance(data, dict):
            h_data = data
        else:
            h_data = data.get(hint_type)
            if h_data is None:
                raise ValueError(f"cannot find data for '{hint_type}'")
        bypass_annotator = h_data.bypass_annotator
        if h_res is not None and not bypass_annotator:
            hint_image = resize_image(hint_image, h_res)
        device = api.device
        use_half = api.use_half
        ht = time.time()
        if need_change_device:
            device = "cuda:0"
            use_half = True
            api.annotators[hint_type].to(device, use_half=True)
        all_annotator_change_device_times.append(time.time() - ht)
        if bypass_annotator:
            o_hint_arr = np.array(hint_image)
        else:
            o_hint_arr = api.get_hint_of(hint_type, hint_image, **h_data.dict())
        ht = time.time()
        if need_change_device:
            api.annotators[hint_type].to("cpu", use_half=False)
            torch.cuda.empty_cache()
        all_annotator_change_device_times.append(time.time() - ht)
        hint_array = cv2.resize(o_hint_arr, (w, h), interpolation=cv2.INTER_LINEAR)
        hint = torch.from_numpy(hint_array)[None].permute(0, 3, 1, 2)
        if use_half:
            hint = hint.half()
        hint = hint.contiguous().to(device) / 255.0
        all_o_hint_arrays.append(o_hint_arr)
        all_hint[hint_type] = hint
    change_annotator_device_time = sum(all_annotator_change_device_times)
    t2 = time.time()
    num_scales = api.m.num_control_scales
    all_scales = {}
    for hint_type in sorted(hint_types):
        h_data = data[hint_type] if isinstance(data, dict) else data
        all_scales[hint_type] = (
            [
                h_data.control_strength * (0.825 ** float(12 - i))
                for i in range(num_scales)
            ]
            if h_data.guess_mode
            else ([h_data.control_strength] * num_scales)
        )
    api.m.control_scales = all_scales
    cond = [common_data.prompt] * common_data.num_samples
    kw = handle_diffusion_model(api, common_data)
    kw["hint"] = all_hint
    kw["hint_start"] = common_data.hint_starts
    kw["max_wh"] = common_data.max_wh
    dt = time.time()
    if need_change_device:
        api.to("cuda:0", use_half=True, no_annotator=True)
    change_diffusion_device_time = time.time() - dt
    # inpainting workaround
    if api.m.unet_kw["in_channels"] == 9:
        if normalized_inpainting_mask is None:
            raise ValueError("`normalized_input_mask` should be provided to inpainting")
        image = Image.fromarray(input_image)
        inpainting_mask = Image.fromarray(to_uint8(normalized_inpainting_mask))
        outs = api.txt2img_inpainting(cond, image, inpainting_mask, **kw)
    elif not common_data.use_img2img:
        kw["size"] = w, h
        outs = api.txt2img(cond, **kw)
    else:
        init_image = cv2.resize(
            input_image,
            (w, h),
            interpolation=cv2.INTER_LINEAR,
        )
        init_image = init_image[None].repeat(common_data.num_samples, axis=0)
        init_image = init_image.transpose([0, 3, 1, 2])
        init_image = init_image.astype(np.float32) / 255.0
        kw["cond"] = cond
        kw["fidelity"] = common_data.fidelity
        init_tensor = torch.from_numpy(init_image)
        if api.use_half:
            init_tensor = init_tensor.half()
        init_tensor = init_tensor.to(api.device)
        outs = api.img2img(init_tensor, **kw)
    dt = time.time()
    api_pool.cleanup(api_key)
    change_diffusion_device_time += time.time() - dt
    outs = 0.5 * (outs + 1.0)
    outs = to_uint8(outs).permute(0, 2, 3, 1).cpu().numpy()
    t3 = time.time()
    results = list(map(resize_to_original, all_o_hint_arrays))
    for i in range(common_data.num_samples):
        results.append(resize_to_original(outs[i]))
    latencies = dict(
        switch=t1 - t0,
        get_hint=t2 - t1 - change_annotator_device_time,
        change_annotator_device=change_annotator_device_time,
        inference=t3 - t2 - change_diffusion_device_time,
        change_diffusion_device=change_diffusion_device_time,
        post_resize=time.time() - t3,
    )
    return results, latencies


async def run_control(
    self: IAlgorithm,
    data: ControlNetModel,
    hint_type: ControlNetHints,
) -> Tuple[List[np.ndarray], Dict[str, float]]:
    self.log_endpoint(data)
    t0 = time.time()
    image, hint_image = await get_images(self, data)
    t1 = time.time()
    results, latencies = apply_control(data, APIs.SD, image, hint_image, hint_type)
    latencies["download"] = t1 - t0
    return results, latencies


def register_control(
    algorithm_model_class: Type[ControlNetModel],
    algorithm_endpoint: str,
    hint_type: ControlNetHints,
) -> None:
    class _(IAlgorithm):
        model_class = algorithm_model_class

        endpoint = algorithm_endpoint

        def initialize(self) -> None:
            register_sd()

        async def run(self, data: algorithm_model_class, *args: Any) -> Response:
            results, latencies = await run_control(self, data, hint_type)
            t0 = time.time()
            content = None if data.return_arrays else np_to_bytes(to_canvas(results))
            t1 = time.time()
            latencies["to_canvas"] = t1 - t0
            self.log_times(latencies)
            if data.return_arrays:
                return results
            return Response(content=content, media_type="image/png")

    IAlgorithm.auto_register()(_)


def register_hint(
    hint_model_class: Type,
    hint_endpoint: str,
    hint_type: ControlNetHints,
) -> None:
    class _Model(hint_model_class, ReturnArraysModel, ImageModel):
        pass

    class _(IAlgorithm):
        model_class = _Model

        endpoint = hint_endpoint

        def initialize(self) -> None:
            register_sd()

        async def run(self, data: _Model, *args: Any) -> Response:
            t0 = time.time()
            image = await self.download_image_with_retry(data.url)
            w, h = image.size
            t1 = time.time()
            m = api_pool.get(APIs.SD)
            t2 = time.time()
            hint_image = np.array(image)
            detect_resolution = getattr(data, "detect_resolution", None)
            if detect_resolution is not None:
                hint_image = resize_image(hint_image, detect_resolution)
            hint = m.get_hint_of(hint_type, hint_image, **data.dict())
            hint = cv2.resize(hint, (w, h), interpolation=cv2.INTER_LINEAR)
            t3 = time.time()
            api_pool.cleanup(APIs.SD)
            self.log_times(
                dict(
                    download=t1 - t0,
                    get_model=t2 - t1,
                    inference=t3 - t2,
                    cleanup=time.time() - t3,
                )
            )
            if data.return_arrays:
                return [hint]
            return Response(content=np_to_bytes(hint), media_type="image/png")

    IAlgorithm.auto_register()(_)


class DetectResolutionModel(BaseModel):
    detect_resolution: int = Field(
        384,
        ge=128,
        le=1024,
        description="Detect resolution.",
    )


class LargeDetectResolutionModel(BaseModel):
    detect_resolution: int = Field(
        512,
        ge=128,
        le=1024,
        description="Detect resolution.",
    )


# ControlNet (depth2image)


class _DepthModel(DetectResolutionModel):
    pass


class DepthModel(_DepthModel, ControlStrengthModel):
    pass


class ControlDepthModel(DepthModel, ControlNetModel):
    pass


# ControlNet (canny2image)


class _CannyModel(LargeDetectResolutionModel):
    low_threshold: int = Field(
        100,
        ge=1,
        le=255,
        description="Low threshold of canny algorithm.",
    )
    high_threshold: int = Field(
        200,
        ge=1,
        le=255,
        description="High threshold of canny algorithm.",
    )


class CannyModel(_CannyModel, ControlStrengthModel):
    pass


class ControlCannyModel(CannyModel, ControlNetModel):
    pass


# ControlNet (pose2image)


class _PoseModel(LargeDetectResolutionModel):
    pass


class PoseModel(_PoseModel, ControlStrengthModel):
    pass


class ControlPoseModel(PoseModel, ControlNetModel):
    pass


# ControlNet (canny2image)


class _MLSDModel(LargeDetectResolutionModel):
    value_threshold: float = Field(
        0.1,
        ge=0.01,
        le=2.0,
        description="Value threshold of mlsd model.",
    )
    distance_threshold: float = Field(
        0.1,
        ge=0.01,
        le=20.0,
        description="Distance threshold of mlsd model.",
    )


class MLSDModel(_MLSDModel, ControlStrengthModel):
    pass


class ControlMLSDModel(MLSDModel, ControlNetModel):
    pass


# register

register_control(ControlDepthModel, control_depth_endpoint, ControlNetHints.DEPTH)
register_control(ControlCannyModel, control_canny_endpoint, ControlNetHints.CANNY)
register_control(ControlPoseModel, control_pose_endpoint, ControlNetHints.POSE)
register_control(ControlMLSDModel, control_mlsd_endpoint, ControlNetHints.MLSD)
register_hint(_DepthModel, control_depth_hint_endpoint, ControlNetHints.DEPTH)
register_hint(_CannyModel, control_canny_hint_endpoint, ControlNetHints.CANNY)
register_hint(_PoseModel, control_pose_hint_endpoint, ControlNetHints.POSE)
register_hint(_MLSDModel, control_mlsd_hint_endpoint, ControlNetHints.MLSD)


__all__ = [
    "control_depth_endpoint",
    "control_canny_endpoint",
    "control_pose_endpoint",
    "control_mlsd_endpoint",
    "control_depth_hint_endpoint",
    "control_canny_hint_endpoint",
    "control_pose_hint_endpoint",
    "control_mlsd_hint_endpoint",
    "ControlDepthModel",
    "ControlCannyModel",
    "ControlPoseModel",
    "ControlMLSDModel",
]



import os
import json
import time
import yaml
import redis
import datetime
import logging.config

from enum import Enum
from kafka import KafkaProducer
from kafka import KafkaAdminClient
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import NamedTuple
from fastapi import FastAPI
from fastapi import Response
from pydantic import Field
from pydantic import BaseModel
from qcloud_cos import CosConfig
from qcloud_cos import CosS3Client
from pkg_resources import get_distribution
from cftool.misc import random_hash
from fastapi.openapi.utils import get_openapi
from fastapi.middleware.cors import CORSMiddleware

from cfclient.utils import get_err_msg
from cfclient.utils import get_responses

from cfcreator import *


app = FastAPI()
root = os.path.dirname(__file__)

origins = [
    "*",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# logging
logging_root = os.path.join(root, "logs", "producer")
os.makedirs(logging_root, exist_ok=True)
with open(os.path.join(root, "config.yml")) as f:
    now = datetime.datetime.now()
    timestamp = now.strftime("%Y-%m-%d_%H-%M-%S-%f")
    log_path = os.path.join(logging_root, f"{timestamp}.log")
    config = yaml.load(f, Loader=yaml.FullLoader)
    config["handlers"]["file"]["filename"] = log_path
    logging.config.dictConfig(config)

excluded_endpoints = {"/health", "/redoc", "/docs", "/openapi.json"}


class EndpointFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        if not record.args:
            return False
        if len(record.args) < 3:
            return False
        if record.args[2] in excluded_endpoints:
            return False
        return True


logging.getLogger("uvicorn.access").addFilter(EndpointFilter())
logging.getLogger("dicttoxml").disabled = True
logging.getLogger("kafka.conn").disabled = True
logging.getLogger("kafka.cluster").disabled = True
logging.getLogger("kafka.coordinator").disabled = True
logging.getLogger("kafka.consumer.subscription_state").disabled = True


# clients
config = CosConfig(Region=REGION, SecretId=SECRET_ID, SecretKey=SECRET_KEY)
cos_client = CosS3Client(config)
redis_client = redis.Redis(**redis_kwargs())
kafka_admin = KafkaAdminClient(bootstrap_servers=kafka_server())
kafka_producer = KafkaProducer(bootstrap_servers=kafka_server())


# schema


DOCS_TITLE = "FastAPI client"
DOCS_VERSION = get_distribution("carefree-client").version
DOCS_DESCRIPTION = (
    "This is a client framework based on FastAPI. "
    "It also supports interacting with Triton Inference Server."
)


def carefree_schema() -> Dict[str, Any]:
    schema = get_openapi(
        title=DOCS_TITLE,
        version=DOCS_VERSION,
        description=DOCS_DESCRIPTION,
        contact={
            "name": "Get Help with this API",
            "email": "syameimaru.saki@gmail.com",
        },
        routes=app.routes,
    )
    app.openapi_schema = schema
    return app.openapi_schema


# health check


class HealthStatus(Enum):
    ALIVE = "alive"


class HealthCheckResponse(BaseModel):
    status: HealthStatus


@app.get("/health", response_model=HealthCheckResponse)
async def health_check() -> HealthCheckResponse:
    return {"status": "alive"}


# get prompt


@app.post("/translate")
@app.post("/get_prompt")
def get_prompt(data: GetPromptModel, response: Response) -> GetPromptResponse:
    inject_headers(response)
    text = data.text
    audit = audit_text(cos_client, text)
    if not audit.safe:
        return GetPromptResponse(text="", success=False, reason=audit.reason)
    return GetPromptResponse(text=text, success=True, reason="")


# kafka & redis


pending_queue_key = get_pending_queue_key()


def get_redis_number(key: str) -> Optional[int]:
    data = redis_client.get(key)
    if data is None:
        return None
    return int(data.decode())  # type: ignore


class ProducerModel(BaseModel):
    task: str
    params: Dict[str, Any]
    notify_url: str = Field("", description="callback url to post to")


class ProducerResponse(BaseModel):
    uid: str


# 30min = timeout
queue_timeout_threshold = 30 * 60


@app.post("/push", responses=get_responses(ProducerResponse))
async def push(data: ProducerModel, response: Response) -> ProducerResponse:
    def get_clean_queue() -> List[str]:
        queue = get_pending_queue()
        queue_size = len(queue)
        # check timeout
        clear_indices = []
        for i, uid in enumerate(queue):
            uid_pack = redis_client.get(uid)
            if uid_pack is None:
                continue
            uid_pack = json.loads(uid_pack)
            create_time = (uid_pack.get("data", {}) or {}).get("create_time", None)
            start_time = (uid_pack.get("data", {}) or {}).get("start_time", None)
            i_cleared = False
            for t in [create_time, start_time]:
                if i_cleared:
                    break
                if t is not None:
                    if time.time() - t >= queue_timeout_threshold:
                        clear_indices.append(i)
                        i_cleared = True
        for idx in clear_indices[::-1]:
            queue.pop(idx)
        # check redundant
        exist = set()
        clear_indices = []
        for i, uid in enumerate(queue):
            if uid not in exist:
                exist.add(uid)
            else:
                clear_indices.append(i)
        for idx in clear_indices[::-1]:
            queue.pop(idx)
        # if pending queue is already updated, ignore the cleanups and return the latest queue
        latest_queue = get_pending_queue()
        if queue_size != len(latest_queue):
            return latest_queue
        # otherwise, return the latest queue
        return queue

    inject_headers(response)
    new_uid = random_hash()
    # append new uid and dump
    queue = get_clean_queue()
    queue.append(new_uid)
    redis_client.set(pending_queue_key, json.dumps(queue))
    redis_client.set(
        new_uid,
        json.dumps(dict(status=Status.PENDING, data=dict(create_time=time.time()))),
    )
    # send to kafka
    data.params["callback_url"] = data.notify_url
    kafka_producer.send(
        kafka_topic(),
        json.dumps(
            dict(
                uid=new_uid,
                task=data.task,
                params=data.params,
            ),
            ensure_ascii=False,
        ).encode("utf-8"),
    )
    return ProducerResponse(uid=new_uid)


class InterruptModel(BaseModel):
    uid_list: List[str]


class InterruptSingleResponse(BaseModel):
    success: bool
    reason: str


class InterruptResponse(BaseModel):
    results: List[InterruptSingleResponse]


@app.post("/interrupt", responses=get_responses(InterruptResponse))
async def interrupt(data: InterruptModel, response: Response) -> InterruptResponse:
    inject_headers(response)
    results = []
    for uid in data.uid_list:
        existing = redis_client.get(uid)
        if existing is None:
            results.append(InterruptSingleResponse(success=False, reason="not found"))
            continue
        existing = json.loads(existing)
        existing_status = existing.get("status")
        if existing_status != Status.PENDING:
            results.append(
                InterruptSingleResponse(
                    success=False,
                    reason=f"not in pending status ({existing_status})",
                )
            )
            continue
        redis_client.set(
            uid,
            json.dumps(dict(status=Status.INTERRUPTED, data=existing.get("data"))),
        )
        results.append(InterruptSingleResponse(success=True, reason=""))
    return results


class ServerStatusResponse(BaseModel):
    is_ready: bool
    num_pending: int
    pending_queue: List[str]


def get_pending_queue() -> List[str]:
    data = redis_client.get(pending_queue_key)
    if data is None:
        return []
    return json.loads(data)


def get_real_pending(queue: List[str]) -> List[str]:
    pending = []
    for uid in queue:
        status = fetch_redis(uid).status
        if status in (Status.FINISHED, Status.EXCEPTION, Status.INTERRUPTED):
            continue
        pending.append(uid)
    return pending


def get_real_lag(queue: List[str]) -> int:
    return len(get_real_pending(queue))


@app.get("/server_status", responses=get_responses(ServerStatusResponse))
async def server_status(response: Response) -> ServerStatusResponse:
    inject_headers(response)
    members = kafka_admin.describe_consumer_groups([kafka_group_id()])[0].members
    real_pending = get_real_pending(get_pending_queue())
    return ServerStatusResponse(
        is_ready=len(members) > 0,
        num_pending=len(real_pending),
        pending_queue=real_pending,
    )


class StatusResponse(BaseModel):
    status: Status
    pending: int
    data: Optional[Any]


class StatusData(NamedTuple):
    status: Status
    data: Optional[Any]


def fetch_redis(uid: str) -> StatusData:
    data = redis_client.get(uid)
    if data is None:
        return StatusData(status=Status.NOT_FOUND, data=None)
    try:
        return StatusData(**json.loads(data))
    except Exception as e:
        return StatusData(status=Status.EXCEPTION, data=dict(reason=get_err_msg(e)))


def _get_status(uid: str) -> StatusResponse:
    record = fetch_redis(uid)
    if record.status != Status.PENDING:
        lag = 0
    else:
        queue = get_pending_queue()
        queue_size = len(queue)
        pop_indices = []
        for i, i_uid in enumerate(queue):
            if fetch_redis(i_uid).status == Status.FINISHED:
                pop_indices.append(i)
        for idx in pop_indices[::-1]:
            queue.pop(idx)
        # update pending queue only if it is not updated during the cleanup
        latest_queue = get_pending_queue()
        if pop_indices and queue_size == len(latest_queue):
            redis_client.set(pending_queue_key, json.dumps(queue))
        try:
            lag = get_real_lag(latest_queue[: latest_queue.index(uid)]) + 1
        except:
            lag = len(latest_queue) + 1
            latest_queue.append(uid)
            redis_client.set(pending_queue_key, json.dumps(latest_queue))
    return StatusResponse(status=record.status, data=record.data, pending=lag)


@app.get("/status/{uid}", responses=get_responses(StatusResponse))
async def get_status(uid: str, response: Response) -> StatusResponse:
    inject_headers(response)
    return _get_status(uid)


class BatchStatusModel(BaseModel):
    uid_list: List[str]


class BatchStatusResponse(BaseModel):
    results: List[StatusResponse]


@app.post("/batch_status", responses=get_responses(BatchStatusModel))
async def get_batch_status(
    data: BatchStatusModel,
    response: Response,
) -> BatchStatusResponse:
    inject_headers(response)
    return BatchStatusResponse(results=[_get_status(uid) for uid in data.uid_list])


# audit


class AuditCallbackModel(BaseModel):
    EventName: str
    JobsDetail: AuditJobsDetailModel


@app.post("/audit_callback")
async def audit_callback(data: AuditCallbackModel) -> None:
    key = data.JobsDetail.Object
    redis_client.expire(key, 3600)
    redis_client.set(
        key,
        json.dumps(data.JobsDetail.dict()),
    )


# events


@app.on_event("startup")
async def startup() -> None:
    pass


@app.on_event("shutdown")
async def shutdown() -> None:
    pass


# schema

app.openapi = carefree_schema


if __name__ == "__main__":
    import uvicorn

    uvicorn.run("interface:app", host="0.0.0.0", port=8989, reload=True)



from typing import List

from .cv import *
from .txt2img import *
from .txt2txt import *
from .img2img import *
from .img2txt import *
from .control import *
from .control_multi import *
from .pipeline import *
from .parameters import Focus


endpoint_to_focuses = {
    # txt2img
    txt2img_sd_endpoint: [
        Focus.ALL,
        Focus.SD,
        Focus.SD_BASE,
        Focus.SD_ANIME,
        Focus.CONTROL,
    ],
    txt2img_sd_inpainting_endpoint: [Focus.ALL, Focus.SD],
    txt2img_sd_outpainting_endpoint: [Focus.ALL, Focus.SD],
    # txt2txt
    txt2txt_prompt_enhance_endpoint: [Focus.ALL, Focus.SYNC],
    # img2img
    img2img_sd_endpoint: [
        Focus.ALL,
        Focus.SD,
        Focus.SD_BASE,
        Focus.SD_ANIME,
        Focus.CONTROL,
    ],
    img2img_semantic2img_endpoint: [Focus.ALL],
    img2img_sr_endpoint: [Focus.ALL, Focus.SYNC],
    img2img_inpainting_endpoint: [Focus.ALL, Focus.SYNC],
    img2img_harmonization_endpoint: [Focus.ALL, Focus.SYNC],
    img2img_sod_endpoint: [Focus.ALL, Focus.SYNC],
    # img2txt
    img2txt_caption_endpoint: [Focus.ALL, Focus.SYNC],
    # cv
    cv_affine_endpoint: [Focus.ALL, Focus.SYNC],
    cv_histogram_match_endpoint: [Focus.ALL, Focus.SYNC],
    # ControlNet
    control_depth_endpoint: [Focus.ALL, Focus.CONTROL],
    control_canny_endpoint: [Focus.ALL, Focus.CONTROL],
    control_pose_endpoint: [Focus.ALL, Focus.CONTROL],
    control_mlsd_endpoint: [Focus.ALL, Focus.CONTROL],
    control_multi_endpoint: [Focus.ALL, Focus.CONTROL],
    control_depth_hint_endpoint: [Focus.ALL, Focus.SYNC, Focus.CONTROL],
    control_canny_hint_endpoint: [Focus.ALL, Focus.SYNC, Focus.CONTROL],
    control_pose_hint_endpoint: [Focus.ALL, Focus.SYNC, Focus.CONTROL],
    control_mlsd_hint_endpoint: [Focus.ALL, Focus.SYNC, Focus.CONTROL],
    # pipeline
    paste_pipeline_endpoint: [Focus.ALL, Focus.SYNC, Focus.PIPELINE],
}


def get_endpoints(focus: Focus) -> List[str]:
    return [e for e, focuses in endpoint_to_focuses.items() if focus in focuses]



import cv2
import time
import torch

import numpy as np

from PIL import Image
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from fastapi import Response
from pydantic import Field
from pydantic import BaseModel
from scipy.interpolate import NearestNDInterpolator
from cftool.cv import to_rgb
from cftool.cv import to_uint8
from cftool.cv import read_image
from cftool.cv import np_to_bytes
from cflearn.api.cv import TranslatorAPI
from cflearn.api.cv.third_party.lama import Config

from .utils import api_pool
from .utils import APIs
from .common import register_sd
from .common import register_esr
from .common import register_lama
from .common import register_hrnet
from .common import register_isnet
from .common import register_semantic
from .common import register_esr_anime
from .common import register_inpainting
from .common import get_sd_from
from .common import handle_diffusion_model
from .common import get_bytes_from_diffusion
from .common import get_bytes_from_translator
from .common import get_normalized_arr_from_diffusion
from .common import IAlgorithm
from .common import ImageModel
from .common import Img2ImgModel
from .common import CallbackModel
from .common import Img2ImgDiffusionModel
from .parameters import verbose
from .parameters import get_focus
from .parameters import Focus


img2img_sd_endpoint = "/img2img/sd"
img2img_sr_endpoint = "/img2img/sr"
img2img_inpainting_endpoint = "/img2img/inpainting"
img2img_semantic2img_endpoint = "/img2img/semantic2img"
img2img_harmonization_endpoint = "/img2img/harmonization"
img2img_sod_endpoint = "/img2img/sod"


# img2img (stable diffusion)


class _Img2ImgSDModel(BaseModel):
    text: str = Field(..., description="The text that we want to handle.")
    fidelity: float = Field(
        0.2,
        ge=0.0,
        le=1.0,
        description="The fidelity of the input image.",
    )
    keep_alpha: bool = Field(
        True,
        description="""
Whether the returned image should keep the alpha-channel of the input image or not.
> If the input image is a sketch image, then `keep_alpha` needs to be False in most of the time.  
""",
    )
    wh: Tuple[int, int] = Field(
        (0, 0),
        description="The output size, `0` means as-is",
    )


class Img2ImgSDModel(Img2ImgDiffusionModel, _Img2ImgSDModel):
    pass


@IAlgorithm.auto_register()
class Img2ImgSD(IAlgorithm):
    model_class = Img2ImgSDModel

    endpoint = img2img_sd_endpoint

    def initialize(self) -> None:
        register_sd()

    async def run(self, data: Img2ImgSDModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.download_image_with_retry(data.url)
        t1 = time.time()
        if not data.keep_alpha:
            image = to_rgb(image)
        w, h = data.wh
        if w > 0 and h > 0:
            image = image.resize((w, h), Image.LANCZOS)
        t2 = time.time()
        m = get_sd_from(data)
        t3 = time.time()
        kwargs = handle_diffusion_model(m, data)
        img_arr = m.img2img(
            image,
            cond=[data.text],
            max_wh=data.max_wh,
            fidelity=data.fidelity,
            anchor=64,
            **kwargs,
        ).numpy()[0]
        content = get_bytes_from_diffusion(img_arr)
        t4 = time.time()
        api_pool.cleanup(APIs.SD)
        self.log_times(
            {
                "download": t1 - t0,
                "preprocess": t2 - t1,
                "get_model": t3 - t2,
                "inference": t4 - t3,
                "cleanup": time.time() - t4,
            }
        )
        return Response(content=content, media_type="image/png")


# super resolution (Real-ESRGAN)


class _Img2ImgSRModel(BaseModel):
    is_anime: bool = Field(
        False,
        description="Whether the input image is an anime image or not.",
    )
    target_w: int = Field(0, description="The target width. 0 means as-is.")
    target_h: int = Field(0, description="The target height. 0 means as-is.")


class Img2ImgSRModel(CallbackModel, _Img2ImgSRModel, Img2ImgModel):
    pass


def apply_sr(
    m: TranslatorAPI,
    image: Image.Image,
    max_wh: int,
    target_w: int,
    target_h: int,
) -> Tuple[np.ndarray, Dict[str, float]]:
    t0 = time.time()
    img_arr = m.sr(image, max_wh=max_wh).numpy()[0]
    img_arr = img_arr.transpose([1, 2, 0])
    t1 = time.time()
    h, w = img_arr.shape[:2]
    if target_w and target_h:
        larger = w * h < target_w * target_h
        img_arr = cv2.resize(
            img_arr,
            (target_w, target_h),
            interpolation=cv2.INTER_LANCZOS4 if larger else cv2.INTER_AREA,
        )
    elif target_w or target_h:
        if target_w:
            k = target_w / w
            target_h = round(h * k)
        else:
            k = target_h / h
            target_w = round(w * k)
        img_arr = cv2.resize(
            img_arr,
            (target_w, target_h),
            interpolation=cv2.INTER_LANCZOS4 if k > 1 else cv2.INTER_AREA,
        )
    return img_arr, dict(inference=t1 - t0, resize=time.time() - t1)


@IAlgorithm.auto_register()
class Img2ImgSR(IAlgorithm):
    model_class = Img2ImgSRModel

    endpoint = img2img_sr_endpoint

    def initialize(self) -> None:
        register_esr()
        register_esr_anime()

    async def run(self, data: Img2ImgSRModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.download_image_with_retry(data.url)
        t1 = time.time()
        api_key = APIs.ESR_ANIME if data.is_anime else APIs.ESR
        m = api_pool.get(api_key)
        t2 = time.time()
        img_arr, latencies = apply_sr(
            m,
            image,
            data.max_wh,
            data.target_w,
            data.target_h,
        )
        t3 = time.time()
        content = get_bytes_from_translator(img_arr, transpose=False)
        t4 = time.time()
        api_pool.cleanup(api_key)
        t5 = time.time()
        latencies.update(
            {
                "download": t1 - t0,
                "get_model": t2 - t1,
                "get_bytes": t4 - t3,
                "cleanup": t5 - t4,
            }
        )
        self.log_times(latencies)
        return Response(content=content, media_type="image/png")


# inpainting (LDM, LaMa)


class InpaintingModels(str, Enum):
    SD = "sd"
    LAMA = "lama"


class Img2ImgInpaintingModel(Img2ImgDiffusionModel):
    model: InpaintingModels = Field(
        InpaintingModels.SD,
        description="The inpainting model that we want to use.",
    )
    use_refine: bool = Field(False, description="Whether should we perform refining.")
    use_pipeline: bool = Field(
        False,
        description="Whether should we perform 'inpainting' + 'refining' in one run.",
    )
    refine_fidelity: float = Field(
        0.2,
        description="""
Refine fidelity used in inpainting.
> Only take effects when `use_refine` / `use_pipeline` is set to True.
""",
    )
    mask_url: str = Field(
        ...,
        description="""
The `cdn` / `cos` url of the user's mask.
> `cos` url from `qcloud` is preferred.
> If empty string is provided, then we will use an empty mask, which means we will simply perform an image-to-image transform.  
""",
    )


@IAlgorithm.auto_register()
class Img2ImgInpainting(IAlgorithm):
    model_class = Img2ImgInpaintingModel

    endpoint = img2img_inpainting_endpoint

    def initialize(self) -> None:
        focus = get_focus()
        self.is_sync = focus == Focus.SYNC
        if not self.is_sync:
            register_inpainting()
        register_lama()

    async def run(self, data: Img2ImgInpaintingModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.download_image_with_retry(data.url)
        mask_url = data.mask_url
        if not mask_url:
            mask = Image.new("L", image.size, color=0)
        else:
            mask = await self.download_image_with_retry(mask_url)
        t1 = time.time()
        model = data.model
        api_key = APIs.LAMA if model == InpaintingModels.LAMA else APIs.INPAINTING
        m = api_pool.get(api_key)
        t2 = time.time()
        if model == InpaintingModels.LAMA:
            cfg = Config()
            image_arr = read_image(
                image,
                None,
                anchor=None,
                to_torch_fmt=False,
            ).image
            mask_arr = read_image(
                mask,
                None,
                anchor=None,
                to_mask=True,
                to_torch_fmt=False,
            ).image
            mask_arr[mask_arr > 0.0] = 1.0
            result = m(image_arr, mask_arr, cfg)
            content = np_to_bytes(result)
        else:
            kwargs = handle_diffusion_model(m, data)
            if not data.use_pipeline:
                refine_fidelity = data.refine_fidelity if data.use_refine else None
                img_arr = m.inpainting(
                    image,
                    mask,
                    max_wh=data.max_wh,
                    refine_fidelity=refine_fidelity,
                    **kwargs,
                ).numpy()[0]
            else:
                img_arr = m.inpainting(
                    image,
                    mask,
                    max_wh=data.max_wh,
                    refine_fidelity=None,
                    **kwargs,
                ).numpy()[0]
                img_arr = get_normalized_arr_from_diffusion(img_arr)
                image = Image.fromarray(to_uint8(img_arr))
                img_arr = m.inpainting(
                    image,
                    mask,
                    max_wh=data.max_wh,
                    refine_fidelity=data.refine_fidelity,
                    **kwargs,
                ).numpy()[0]
            content = get_bytes_from_diffusion(img_arr)
        t3 = time.time()
        api_pool.cleanup(api_key)
        self.log_times(
            {
                "download": t1 - t0,
                "get_model": t2 - t1,
                "inference": t3 - t2,
                "cleanup": time.time() - t3,
            }
        )
        return Response(content=content, media_type="image/png")


# semantic2img (LDM)


class Img2ImgSemantic2ImgModel(Img2ImgDiffusionModel):
    color2label: Dict[str, int] = Field(
        ...,
        description="""
Mapping of color -> (semantic) label.
> The color should be of `rgb(r,g,b)` format.
""",
    )
    keep_alpha: bool = Field(
        False,
        description="Whether the returned image should keep the alpha-channel of the input image or not.",
    )


def color2rgb(color: str) -> List[int]:
    if not color.startswith("rgb(") or not color.endswith(")"):
        raise ValueError("`color` should be of `rgb(r,g,b)` format")
    return [int(n.strip()) for n in color[4:-1].split(",")]


@IAlgorithm.auto_register()
class Img2ImgSemantic2Img(IAlgorithm):
    model_class = Img2ImgSemantic2ImgModel

    endpoint = img2img_semantic2img_endpoint

    def initialize(self) -> None:
        register_semantic()

    async def run(self, data: Img2ImgSemantic2ImgModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        raw_semantic = await self.download_image_with_retry(data.url)
        t1 = time.time()
        w, h = raw_semantic.size
        raw_arr = np.array(raw_semantic)
        alpha = None
        valid_mask = None
        valid_mask_ravel = None
        # handle alpha
        if raw_arr.shape[-1] == 4:
            alpha = raw_arr[..., -1]
            raw_arr = raw_arr[..., :3]
            valid_mask = alpha > 0
            valid_mask_ravel = valid_mask.ravel()
        # get nearest color
        raw_arr_flat = raw_arr.reshape([h * w, 1, 3])
        if valid_mask_ravel is not None:
            raw_arr_flat = raw_arr_flat[valid_mask_ravel]
        raw_arr_flat = raw_arr_flat.astype(np.int16)
        colors = sorted(data.color2label)
        rgbs = np.array(list(map(color2rgb, colors)), np.int16).reshape([1, -1, 3])
        diff = np.abs(raw_arr_flat - rgbs).mean(2)
        indices = np.argmin(diff, axis=1)
        # diffusion has no `unlabeled` label, so it should be COCO.label - 1
        labels = np.array([data.color2label[color] - 1 for color in colors], np.uint8)
        arr_labels = labels[indices]
        if valid_mask_ravel is None:
            semantic_arr = arr_labels
        else:
            semantic_arr = np.zeros([h, w], np.uint8).ravel()
            semantic_arr[valid_mask_ravel] = arr_labels
        # nearest interpolation
        t2 = time.time()
        if valid_mask is not None and valid_mask_ravel is not None:
            to_coordinates = lambda mask: np.array(np.nonzero(mask)).T
            valid_coordinates = to_coordinates(valid_mask)
            interpolator = NearestNDInterpolator(valid_coordinates, arr_labels)
            invalid_mask = ~valid_mask
            invalid_coordinates = to_coordinates(invalid_mask)
            semantic_arr[invalid_mask.ravel()] = interpolator(invalid_coordinates)
        # gather
        semantic_arr = semantic_arr.reshape([h, w])
        semantic = Image.fromarray(semantic_arr)
        t3 = time.time()
        m = api_pool.get(APIs.SEMANTIC)
        t4 = time.time()
        if not data.keep_alpha:
            alpha = None
        elif alpha is not None:
            alpha = alpha[None, None].astype(np.float32) / 255.0
        img_arr = m.semantic2img(
            semantic,
            alpha=alpha,
            max_wh=data.max_wh,
            verbose=verbose(),
            seed=data.seed,
        ).numpy()[0]
        content = get_bytes_from_diffusion(img_arr)
        t5 = time.time()
        api_pool.cleanup(APIs.SEMANTIC)
        self.log_times(
            {
                "download": t1 - t0,
                "preprocess": t2 - t1,
                "interpolation": t3 - t2,
                "get_model": t4 - t3,
                "inference": t5 - t4,
                "cleanup": time.time() - t5,
            }
        )
        return Response(content=content, media_type="image/png")


# image harmonization (hrnet)


def apply_harmonization(
    max_wh: int,
    strength: float,
    raw_image: np.ndarray,
    normalized_mask: np.ndarray,
) -> Tuple[np.ndarray, Dict[str, float]]:
    t0 = time.time()
    m = api_pool.get(APIs.HRNET)
    t1 = time.time()
    h, w = raw_image.shape[:2]
    scale = max_wh**2 / (w * h)
    if scale >= 1.0:
        scaled_image = raw_image
        scaled_mask = normalized_mask
    else:
        scaled_w = round(w * scale)
        scaled_h = round(h * scale)
        scaled_image = cv2.resize(raw_image, (scaled_w, scaled_h))
        scaled_mask = cv2.resize(normalized_mask, (scaled_w, scaled_h))
    result = m.predict(scaled_image, scaled_mask)
    if scale < 1.0:
        result = cv2.resize(result, (w, h))
    if strength != 1.0:
        raw_image = raw_image.astype(np.float32)
        result = result.astype(np.float32)
        result = result * strength + raw_image * (1.0 - strength)
        result = (np.clip(result, 0.0, 255.0)).astype(np.uint8)
    t2 = time.time()
    api_pool.cleanup(APIs.HRNET)
    latencies = {
        "get_model": t1 - t0,
        "inference": t2 - t1,
        "cleanup": time.time() - t2,
    }
    return result, latencies


class Img2ImgHarmonizationModel(ImageModel):
    mask_url: str = Field(
        ...,
        description="The `cdn` / `cos` url of the harmonization mask. (`cos` url is preferred)",
    )
    strength: float = Field(1.0, description="Strength of the harmonization process.")
    harmonization_max_wh: int = Field(
        2048,
        description="max_wh for the harmonization inputs.",
    )


@IAlgorithm.auto_register()
class Img2ImgHarmonization(IAlgorithm):
    model_class = Img2ImgHarmonizationModel

    endpoint = img2img_harmonization_endpoint

    def initialize(self) -> None:
        register_hrnet()

    async def run(self, data: Img2ImgHarmonizationModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.download_image_with_retry(data.url)
        mask = await self.download_image_with_retry(data.mask_url)
        t1 = time.time()
        result, latencies = apply_harmonization(
            data.harmonization_max_wh,
            data.strength,
            read_image(
                image,
                None,
                anchor=None,
                normalize=False,
                to_torch_fmt=False,
            ).image,
            read_image(
                mask,
                None,
                anchor=None,
                to_mask=True,
                to_torch_fmt=False,
            ).image,
        )
        latencies["download"] = t1 - t0
        self.log_times(latencies)
        return Response(content=np_to_bytes(result), media_type="image/png")


# salient object detection (isnet)


@IAlgorithm.auto_register()
class Img2ImgSOD(IAlgorithm):
    model_class = ImageModel

    endpoint = img2img_sod_endpoint

    def initialize(self) -> None:
        register_isnet()

    async def run(self, data: ImageModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.download_image_with_retry(data.url)
        t1 = time.time()
        m = api_pool.get(APIs.ISNET)
        t2 = time.time()
        alpha = to_uint8(m.segment(image))
        content = np_to_bytes(alpha)
        t3 = time.time()
        api_pool.cleanup(APIs.ISNET)
        self.log_times(
            {
                "download": t1 - t0,
                "get_model": t2 - t1,
                "inference": t3 - t2,
                "cleanup": time.time() - t3,
            }
        )
        return Response(content=content, media_type="image/png")


__all__ = [
    "img2img_sd_endpoint",
    "img2img_sr_endpoint",
    "img2img_inpainting_endpoint",
    "img2img_semantic2img_endpoint",
    "img2img_harmonization_endpoint",
    "img2img_sod_endpoint",
    "Img2ImgSDModel",
    "Img2ImgSRModel",
    "Img2ImgInpaintingModel",
    "Img2ImgSemantic2ImgModel",
    "Img2ImgSD",
    "Img2ImgSR",
    "Img2ImgInpainting",
    "Img2ImgSemantic2Img",
    "Img2ImgHarmonizationModel",
    "Img2ImgHarmonization",
    "Img2ImgSOD",
]



# `kafka` sdk is used to access to the server launched by `uvicorn apis.kafka.producer:app --host 0.0.0.0 --port 8123`

import json
import time
import requests

from typing import Any
from typing import Dict
from typing import Callable
from typing import Optional
from cfclient.utils import get
from cfclient.utils import post
from cfcreator.common import endpoint2algorithm
from cfcreator.common import Status

from .utils import *


TCallback = Callable[[Dict[str, Any]], None]


async def push(host: str, endpoint: str, params: Dict[str, Any]) -> str:
    url = f"{host}/push"
    data = dict(task=endpoint2algorithm(endpoint), params=params)
    async with get_http_session() as session:
        res = await post(url, data, session)
        return res["uid"]


def push_sync(host: str, endpoint: str, params: Dict[str, Any]) -> str:
    url = f"{host}/push"
    data = dict(task=endpoint2algorithm(endpoint), params=params)
    return requests.post(url, json=data).json()["uid"]


async def poll(
    host: str,
    uid: str,
    *,
    callback: Optional[TCallback] = None,
) -> Dict[str, Any]:
    url = f"{host}/status/{uid}"
    async with get_http_session() as session:
        while True:
            res = json.loads(await get(url, session))
            if callback is not None:
                callback(res)
            if res["status"] == Status.PENDING:
                time.sleep(res["pending"])
            elif res["status"] == Status.WORKING:
                time.sleep(1)
            else:
                return res


def poll_sync(
    host: str,
    uid: str,
    *,
    callback: Optional[TCallback] = None,
) -> Dict[str, Any]:
    while True:
        res = requests.get(f"{host}/status/{uid}").json()
        if callback is not None:
            callback(res)
        if res["status"] == Status.PENDING:
            time.sleep(res["pending"])
        elif res["status"] == Status.WORKING:
            time.sleep(1)
        else:
            return res


__all__ = [
    "push",
    "poll",
    "push_sync",
    "poll_sync",
]



import cv2
import time

import numpy as np

from typing import Any
from fastapi import Response
from pydantic import Field
from pydantic import BaseModel
from cftool.cv import to_uint8
from cftool.cv import np_to_bytes
from cftool.cv import ImageProcessor
from cfclient.models import ImageModel

from .common import IAlgorithm


cv_affine_endpoint = "/cv/affine"
cv_histogram_match_endpoint = "/cv/hist_match"


def affine(
    array: np.ndarray,
    a: float,
    b: float,
    c: float,
    d: float,
    e: float,
    f: float,
    w: int,
    h: int,
) -> np.ndarray:
    matrix = np.array([[a, c, e], [b, d, f]], np.float32)
    return cv2.warpAffine(array, matrix, [w, h])


class BaseAffineModel(BaseModel):
    a: float = Field(..., description="`a` of the affine matrix")
    b: float = Field(..., description="`b` of the affine matrix")
    c: float = Field(..., description="`c` of the affine matrix")
    d: float = Field(..., description="`d` of the affine matrix")
    e: float = Field(..., description="`e` of the affine matrix")
    f: float = Field(..., description="`f` of the affine matrix")


class AffineModel(BaseAffineModel, ImageModel):
    w: int = Field(..., description="width of the output image")
    h: int = Field(..., description="width of the output image")


@IAlgorithm.auto_register()
class Affine(IAlgorithm):
    model_class = AffineModel

    endpoint = cv_affine_endpoint

    def initialize(self) -> None:
        pass

    async def run(self, data: AffineModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.download_image_with_retry(data.url)
        t1 = time.time()
        output = affine(
            np.array(image),
            data.a,
            data.b,
            data.c,
            data.d,
            data.e,
            data.f,
            data.w,
            data.h,
        )
        t2 = time.time()
        content = np_to_bytes(output)
        self.log_times(
            {
                "download": t1 - t0,
                "preprocess": t2 - t1,
                "to_bytes": time.time() - t2,
            }
        )
        return Response(content=content, media_type="image/png")


class HistogramMatchModel(ImageModel):
    bg_url: str = Field(..., description="The `cdn` / `cos` url of the background.")
    use_hsv: bool = Field(False, description="Whether use the HSV space to match.")
    strength: float = Field(1.0, description="Strength of the matching.")


@IAlgorithm.auto_register()
class HistogramMatch(IAlgorithm):
    model_class = HistogramMatchModel

    endpoint = cv_histogram_match_endpoint

    def initialize(self) -> None:
        pass

    async def run(self, data: HistogramMatchModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        image = await self.download_image_with_retry(data.url)
        bg = await self.download_image_with_retry(data.bg_url)
        t1 = time.time()
        to_normalized = lambda im: np.array(im).astype(np.float32) / 255.0
        rgba, bg_arr = map(to_normalized, [image, bg])
        rgb, alpha = rgba[..., :3], rgba[..., -1:]
        merged = rgb * alpha + bg_arr * (1.0 - alpha)
        merged, bg_arr = map(to_uint8, [merged, bg_arr])
        t2 = time.time()
        if data.use_hsv:
            merged = cv2.cvtColor(merged, cv2.COLOR_RGB2HSV)
            bg_arr = cv2.cvtColor(bg_arr, cv2.COLOR_RGB2HSV)
        adjusted = ImageProcessor.match_histograms(
            merged,
            bg_arr,
            alpha[..., 0] > 0,
            strength=data.strength,
        )
        if data.use_hsv:
            adjusted = cv2.cvtColor(adjusted, cv2.COLOR_HSV2RGB)
        t3 = time.time()

        from PIL import Image

        Image.fromarray(adjusted).save("adjusted.png")

        content = np_to_bytes(np.zeros([64, 64]))
        # content = np_to_bytes(adjusted)
        # content = np_to_bytes(merged)
        self.log_times(
            {
                "download": t1 - t0,
                "preprocess": t2 - t1,
                "calculation": t3 - t2,
                "to_bytes": time.time() - t3,
            }
        )
        return Response(content=content, media_type="image/png")


__all__ = [
    "cv_affine_endpoint",
    "cv_histogram_match_endpoint",
    "BaseAffineModel",
    "AffineModel",
    "HistogramMatchModel",
    "Affine",
    "HistogramMatch",
]



from setuptools import setup, find_packages

VERSION = "0.2.1"
PACKAGE_NAME = "carefree-creator"

DESCRIPTION = "An AI-powered creator for everyone."
with open("README.md", encoding="utf-8") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name=PACKAGE_NAME,
    version=VERSION,
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    entry_points={"console_scripts": ["cfcreator = cfcreator.cli:main"]},
    install_requires=[
        "click>=8.1.3",
        "fastapi==0.88.0",
        "carefree-client>=0.1.9",
        "carefree-learn[cv_full]>=0.4.1",
    ],
    extras_require={
        "kafka": [
            "kafka-python",
            "redis[hiredis]",
            "cos-python-sdk-v5",
        ]
    },
    author="carefree0910",
    author_email="syameimaru.saki@gmail.com",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords="python carefree-learn PyTorch",
)



import os
import json
import yaml
import torch
import logging
import datetime
import logging.config

from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from fastapi import FastAPI
from fastapi import Response
from pydantic import BaseModel
from pkg_resources import get_distribution
from cftool.array import tensor_dict_type
from fastapi.openapi.utils import get_openapi
from fastapi.middleware.cors import CORSMiddleware

from cfclient.models import *
from cfclient.core import HttpClient
from cfclient.core import TritonClient
from cfclient.utils import get_responses
from cfclient.utils import run_algorithm
from cfclient.utils import get_image_response_kwargs

from cfcreator import *
from cfcreator.utils import api_pool


app = FastAPI()
root = os.path.dirname(__file__)

constants = dict(
    triton_host=None,
    triton_port=8000,
    model_root=os.path.join(root, "models"),
    token_root=os.path.join(root, "tokens"),
)

origins = [
    "*",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

# logging
logging_root = os.path.join(root, "logs")
os.makedirs(logging_root, exist_ok=True)
with open(os.path.join(root, "config.yml")) as f:
    now = datetime.datetime.now()
    timestamp = now.strftime("%Y-%m-%d_%H-%M-%S-%f")
    log_path = os.path.join(logging_root, f"{timestamp}.log")
    config = yaml.load(f, Loader=yaml.FullLoader)
    config["handlers"]["file"]["filename"] = log_path
    logging.config.dictConfig(config)

excluded_endpoints = {"/health", "/redoc", "/docs", "/openapi.json"}


class EndpointFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        if not record.args:
            return False
        if len(record.args) < 3:
            return False
        if record.args[2] in excluded_endpoints:
            return False
        return True


logging.getLogger("uvicorn.access").addFilter(EndpointFilter())

# clients
## http client
http_client = HttpClient()
## triton client
triton_host = constants["triton_host"]
if triton_host is None:
    triton_client = None
else:
    triton_client = TritonClient(url=f"{triton_host}:{constants['triton_port']}")
## collect
clients = dict(
    http=http_client,
    triton=triton_client,
)


# algorithms
all_algorithms: Dict[str, AlgorithmBase] = {
    k: v(clients) for k, v in algorithms.items()
}


# schema


DOCS_TITLE = "FastAPI client"
DOCS_VERSION = get_distribution("carefree-client").version
DOCS_DESCRIPTION = (
    "This is a client framework based on FastAPI. "
    "It also supports interacting with Triton Inference Server."
)


def carefree_schema() -> Dict[str, Any]:
    schema = get_openapi(
        title=DOCS_TITLE,
        version=DOCS_VERSION,
        description=DOCS_DESCRIPTION,
        contact={
            "name": "Get Help with this API",
            "email": "syameimaru.saki@gmail.com",
        },
        routes=app.routes,
    )
    app.openapi_schema = schema
    return app.openapi_schema


# health check


class HealthStatus(Enum):
    ALIVE = "alive"


class HealthCheckResponse(BaseModel):
    status: HealthStatus


@app.get("/health", response_model=HealthCheckResponse)
async def health_check() -> HealthCheckResponse:
    return {"status": "alive"}


# demo


@app.post(demo_hello_endpoint, responses=get_responses(HelloResponse))
async def hello(data: HelloModel) -> HelloResponse:
    return await run_algorithm(all_algorithms["demo.hello"], data)


# get prompt


@app.post("/translate", responses=get_responses(GetPromptResponse))
@app.post("/get_prompt", responses=get_responses(GetPromptResponse))
def get_prompt(data: GetPromptModel) -> GetPromptResponse:
    return GetPromptResponse(text=data.text, success=True, reason="")


# switch local checkpoint


class ModelRootResponse(BaseModel):
    root: str


class SwitchCheckpointModel(BaseModel):
    key: str
    model: str


class SwitchCheckpointResponse(BaseModel):
    success: bool
    reason: str


class AvailableVersions(BaseModel):
    versions: List[str]


class AvailableModels(BaseModel):
    models: List[str]


class SwitchCheckpointRootModel(BaseModel):
    root: str


class SwitchCheckpointRootResponse(BaseModel):
    success: bool
    reason: str


class ResetCheckpointModel(BaseModel):
    version: str


class ResetCheckpointResponse(BaseModel):
    success: bool
    reason: str


def _get_available_local_models(root: str) -> List[str]:
    if not os.path.isdir(root):
        return []
    return list(
        filter(
            lambda file: file.endswith(".pt") or file.endswith(".ckpt"),
            os.listdir(root),
        )
    )


@app.post("/model_root", responses=get_responses(GetPromptResponse))
def get_model_root() -> ModelRootResponse:
    return ModelRootResponse(root=constants["model_root"])


@app.post("/available_models", responses=get_responses(GetPromptResponse))
def get_available_local_models() -> AvailableModels:
    return AvailableModels(models=_get_available_local_models(constants["model_root"]))


@app.post("/switch_root", responses=get_responses(SwitchCheckpointRootResponse))
def switch_root(data: SwitchCheckpointRootModel) -> SwitchCheckpointRootResponse:
    if not _get_available_local_models(data.root):
        return SwitchCheckpointRootResponse(
            success=False,
            reason=f"cannot find any checkpoints under '{data.root}'",
        )
    constants["model_root"] = data.root
    return SwitchCheckpointRootResponse(success=True, reason="")


# inject custom tokens


custom_embeddings: tensor_dict_type = {}


def _inject_custom_tokens(root: str) -> tensor_dict_type:
    local_customs: tensor_dict_type = {}
    if not os.path.isdir(root):
        return local_customs
    for file in os.listdir(root):
        try:
            path = os.path.join(root, file)
            d = torch.load(path, map_location="cpu")
            local_customs.update({k: v.tolist() for k, v in d.items()})
        except:
            continue
    if local_customs:
        print(f"> Following tokens are loaded: {', '.join(sorted(local_customs))}")
        custom_embeddings.update(local_customs)
    return local_customs


class InjectCustomTokenModel(BaseModel):
    root: str


class InjectCustomTokenResponse(BaseModel):
    success: bool
    reason: str


@app.post("/inject_tokens", responses=get_responses(InjectCustomTokenResponse))
def inject_custom_tokens(data: InjectCustomTokenModel) -> InjectCustomTokenResponse:
    if not _inject_custom_tokens(data.root):
        return InjectCustomTokenResponse(
            success=False,
            reason=f"cannot find any tokens under '{data.root}'",
        )
    return InjectCustomTokenResponse(success=True, reason="")


# meta


env_opt_json = os.environ.get(OPT_ENV_KEY)
if env_opt_json is not None:
    OPT.update(json.loads(env_opt_json))
focus = get_focus()
registered_algorithms = set()
api_pool.update_limit()


def register_endpoint(endpoint: str) -> None:
    name = endpoint[1:].replace("/", "_")
    algorithm_name = endpoint2algorithm(endpoint)
    algorithm: IAlgorithm = all_algorithms[algorithm_name]
    registered_algorithms.add(algorithm_name)
    data_model = algorithm.model_class
    if algorithm.response_model_class is None:
        response_model = Response
        post_kwargs = get_image_response_kwargs()
    else:
        response_model = algorithm.response_model_class
        post_kwargs = dict(responses=get_responses(response_model))

    @app.post(endpoint, **post_kwargs, name=name)
    async def _(data: data_model) -> response_model:
        if (
            isinstance(data, (Txt2ImgSDModel, Img2ImgSDModel))
            and not data.custom_embeddings
        ):
            data.custom_embeddings = {
                token: embedding
                for token, embedding in custom_embeddings.items()
                if token in data.text or token in data.negative_prompt
            }
        return await run_algorithm(algorithm, data)


for endpoint in get_endpoints(focus):
    register_endpoint(endpoint)


# events


@app.on_event("startup")
async def startup() -> None:
    http_client.start()
    OPT["use_cos"] = False
    for k, v in all_algorithms.items():
        if k in registered_algorithms:
            v.initialize()
    _inject_custom_tokens(constants["token_root"])
    print("> Server is Ready!")


@app.on_event("shutdown")
async def shutdown() -> None:
    await http_client.stop()


# schema

app.openapi = carefree_schema


if __name__ == "__main__":
    import uvicorn

    uvicorn.run("interface:app", host="0.0.0.0", port=8989, reload=True)



import time

import numpy as np

from PIL import Image
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from fastapi import Response
from pydantic import Field
from pydantic import BaseModel
from cftool.cv import to_rgb
from cftool.cv import to_uint8
from cftool.cv import np_to_bytes
from cfclient.models.core import ImageModel

from .cv import affine
from .cv import BaseAffineModel
from .utils import to_canvas
from .common import IAlgorithm
from .common import ReturnArraysModel


paste_pipeline_endpoint = "/pipeline/paste"


def get_response(data: ReturnArraysModel, results: List[np.ndarray]) -> Any:
    if data.return_arrays:
        return results
    return Response(content=np_to_bytes(to_canvas(results)), media_type="image/png")


# paste pipeline


def paste(
    original_fg: Image.Image,
    original_bg: Image.Image,
    a: float,
    b: float,
    c: float,
    d: float,
    e: float,
    f: float,
) -> Tuple[Dict[str, np.ndarray], Dict[str, Any]]:
    t0 = time.time()
    if original_fg.mode != "RGBA":
        original_fg = original_fg.convert("RGBA")
    original_w, original_h = original_bg.size
    affined_fg_array = affine(
        np.array(original_fg),
        a,
        b,
        c,
        d,
        e,
        f,
        original_w,
        original_h,
    )
    t1 = time.time()
    rgb = affined_fg_array[..., :3].astype(np.float32) / 255.0
    mask = affined_fg_array[..., -1:].astype(np.float32) / 255.0
    original_bg_array = np.array(to_rgb(original_bg)).astype(np.float32) / 255.0
    merged = rgb * mask + original_bg_array * (1.0 - mask)
    results = dict(rgb=rgb, mask=mask, merged=to_uint8(merged))
    latencies = {"affine": t1 - t0, "merge": time.time() - t1}
    return results, latencies


class _PastePipelineModel(BaseModel):
    bg_url: str = Field(
        ...,
        description="""
The `cdn` / `cos` url of the background's image.
> `cos` url from `qcloud` is preferred.
""",
    )


class PastePipelineModel(
    ReturnArraysModel,
    BaseAffineModel,
    _PastePipelineModel,
    ImageModel,
):
    pass


@IAlgorithm.auto_register()
class PastePipeline(IAlgorithm):
    model_class = PastePipelineModel

    endpoint = paste_pipeline_endpoint

    def initialize(self) -> None:
        pass

    async def run(self, data: PastePipelineModel, *args: Any) -> Response:
        self.log_endpoint(data)
        t0 = time.time()
        original_fg = await self.download_image_with_retry(data.url)
        original_bg = await self.download_image_with_retry(data.bg_url)
        t1 = time.time()
        results, latencies = paste(
            original_fg,
            original_bg,
            data.a,
            data.b,
            data.c,
            data.d,
            data.e,
            data.f,
        )
        latencies["download"] = t1 - t0
        self.log_times(latencies)
        return get_response(data, [results["merged"]])


__all__ = [
    "paste_pipeline_endpoint",
    "PastePipelineModel",
    "PastePipeline",
]



import os
import json
import torch
import cflearn

import numpy as np

from abc import ABCMeta
from PIL import Image
from enum import Enum
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Callable
from typing import Optional
from pydantic import Field
from pydantic import BaseModel
from functools import partial
from cftool.cv import np_to_bytes
from cfclient.models import TextModel
from cfclient.models import ImageModel
from cfclient.models import AlgorithmBase
from cflearn.api.cv import SDVersions
from cflearn.api.cv import DiffusionAPI
from cflearn.api.cv import TranslatorAPI
from cflearn.api.cv import ImageHarmonizationAPI
from cflearn.api.cv import ControlledDiffusionAPI
from cflearn.misc.toolkit import _get_file_size
from cflearn.api.cv.third_party.blip import BLIPAPI
from cflearn.api.cv.third_party.lama import LaMa
from cflearn.api.cv.third_party.isnet import ISNetAPI
from cflearn.api.cv.third_party.prompt import PromptEnhanceAPI

from .cos import download_with_retry
from .cos import download_image_with_retry
from .utils import api_pool
from .utils import APIs
from .parameters import verbose
from .parameters import get_focus
from .parameters import pool_limit
from .parameters import Focus


class ExternalVersions(str, Enum):
    """
    Specify external SD weights that need to be loaded.
    * these weights should be placed under ~/.cache/external/ folder.
    * file name should be {version}.ckpt

    Example
    -------
    class ExternalVersions(str, Enum):
        MY_FANCY_MODEL = "my_fancy_model"

    then you can place your model at ~/.cache/external/my_fancy_model.ckpt,
    after which you can specify `my_fancy_model` as the `version` parameter!
    """


def merge_enums(*enums: Enum) -> Enum:
    members: Dict[str, str] = {}
    for e in enums:
        for name, member in e.__members__.items():
            members[name] = member.value
    return Enum("MergedVersions", members, type=str)


MergedVersions = merge_enums(SDVersions, ExternalVersions)


def _get(init_fn: Callable, init_to_cpu: bool) -> Any:
    if init_to_cpu:
        return init_fn()
    return init_fn("cuda:0", use_half=True)


def init_sd(init_to_cpu: bool) -> ControlledDiffusionAPI:
    version = MergedVersions.v1_5
    init_fn = partial(ControlledDiffusionAPI.from_sd_version, version)
    m: ControlledDiffusionAPI = _get(init_fn, init_to_cpu)
    focus = get_focus()
    m.sd_weights.limit = pool_limit()
    m.current_sd_version = version
    targets = []
    common = Focus.ALL, Focus.SD, Focus.CONTROL, Focus.PIPELINE
    if focus in common + (Focus.SD_BASE,):
        targets.append(MergedVersions.v1_5)
    if focus in common + (Focus.SD_ANIME,):
        targets.append(MergedVersions.ANIME)
        targets.append(MergedVersions.DREAMLIKE)
        targets.append(MergedVersions.ANIME_ANYTHING)
        targets.append(MergedVersions.ANIME_HYBRID)
        targets.append(MergedVersions.ANIME_GUOFENG)
        targets.append(MergedVersions.ANIME_ORANGE)
    print(f"> preparing sd weights ({', '.join(targets)}) (focus={focus})")
    m.prepare_sd(targets)
    # when focus is SYNC, `init_sd` is called because we need to expose `control_hint`
    # endpoints. However, `sd` itself will never be used, so we can skip some stuffs
    if focus == Focus.SYNC:
        print("> prepare ControlNet Annotators")
        for hint in m.control_defaults:
            m.prepare_annotator(hint)
    else:
        print("> handling external weights")
        external_dir = os.path.join(os.path.expanduser("~"), ".cache", "external")
        os.makedirs(external_dir, exist_ok=True)
        converted_sizes_path = os.path.join(external_dir, "sizes.json")
        sizes: Dict[str, int]
        if not os.path.isfile(converted_sizes_path):
            sizes = {}
        else:
            with open(converted_sizes_path, "r") as f:
                sizes = json.load(f)
        for version in ExternalVersions:
            print(f">> handling {version}")
            converted_path = os.path.join(external_dir, f"{version}_converted.pt")
            v_size = sizes.get(version)
            f_size = (
                None
                if not os.path.isfile(converted_path)
                else _get_file_size(converted_path)
            )
            if f_size is None or v_size != f_size:
                if f_size is not None:
                    print(f">> {version} has been converted but size mismatch")
                print(f">> converting {version}")
                model_path = os.path.join(external_dir, f"{version}.ckpt")
                d = cflearn.scripts.sd.convert(model_path, m, load=False)
                torch.save(d, converted_path)
                sizes[version] = _get_file_size(converted_path)
            m.sd_weights.register(version, converted_path)
        with open(converted_sizes_path, "w") as f:
            json.dump(sizes, f)
        print("> prepare ControlNet weights")
        m.prepare_control_defaults()
        print("> prepare ControlNet Annotators")
        m.prepare_annotators()
        print("> warmup ControlNet")
        m.switch_control(*m.available_control_hints)
    # return
    return m


def init_sd_inpainting(init_to_cpu: bool) -> ControlledDiffusionAPI:
    register_sd()
    sd: ControlledDiffusionAPI = api_pool.get(APIs.SD)
    init_fn = ControlledDiffusionAPI.from_sd_inpainting
    m: ControlledDiffusionAPI = _get(init_fn, init_to_cpu)
    m.annotators = sd.annotators
    m.controlnet_weights = sd.controlnet_weights
    m.current_sd_version = MergedVersions.v1_5
    m.switch_control(*m.available_control_hints)
    return m


def register_sd() -> None:
    api_pool.register(APIs.SD, init_sd)


def register_sd_inpainting() -> None:
    api_pool.register(APIs.SD_INPAINTING, init_sd_inpainting)


def register_esr() -> None:
    api_pool.register(
        APIs.ESR,
        lambda init_to_cpu: _get(TranslatorAPI.from_esr, init_to_cpu),
    )


def register_esr_anime() -> None:
    api_pool.register(
        APIs.ESR_ANIME,
        lambda init_to_cpu: _get(TranslatorAPI.from_esr_anime, init_to_cpu),
    )


def register_inpainting() -> None:
    api_pool.register(
        APIs.INPAINTING,
        lambda init_to_cpu: _get(DiffusionAPI.from_inpainting, init_to_cpu),
    )


def register_lama() -> None:
    api_pool.register(
        APIs.LAMA,
        lambda init_to_cpu: _get(LaMa, init_to_cpu),
    )


def register_semantic() -> None:
    api_pool.register(
        APIs.SEMANTIC,
        lambda init_to_cpu: _get(DiffusionAPI.from_semantic, init_to_cpu),
    )


def register_hrnet() -> None:
    api_pool.register(
        APIs.HRNET,
        lambda init_to_cpu: _get(ImageHarmonizationAPI, init_to_cpu),
    )


def register_isnet() -> None:
    api_pool.register(
        APIs.ISNET,
        lambda init_to_cpu: _get(ISNetAPI, init_to_cpu),
    )


def register_blip() -> None:
    api_pool.register(
        APIs.BLIP,
        lambda init_to_cpu: _get(BLIPAPI, init_to_cpu),
    )


def register_prompt_enhance() -> None:
    api_pool.register(
        APIs.PROMPT_ENHANCE,
        lambda init_to_cpu: _get(PromptEnhanceAPI, init_to_cpu),
    )


def get_bytes_from_translator(img_arr: np.ndarray, *, transpose: bool = True) -> bytes:
    if transpose:
        img_arr = img_arr.transpose([1, 2, 0])
    return np_to_bytes(img_arr)


def get_normalized_arr_from_diffusion(img_arr: np.ndarray) -> np.ndarray:
    img_arr = 0.5 * (img_arr + 1.0)
    img_arr = img_arr.transpose([1, 2, 0])
    return img_arr


def get_bytes_from_diffusion(img_arr: np.ndarray) -> bytes:
    return np_to_bytes(get_normalized_arr_from_diffusion(img_arr))


# API models


class CallbackModel(BaseModel):
    callback_url: str = Field("", description="callback url to post to")


class MaxWHModel(BaseModel):
    max_wh: int = Field(1024, description="The maximum resolution.")


class VariationModel(BaseModel):
    seed: int = Field(..., description="Seed of the variation.")
    strength: float = Field(
        ...,
        ge=0.0,
        le=1.0,
        description="Strength of the variation.",
    )


class SDSamplers(str, Enum):
    DDIM = "ddim"
    PLMS = "plms"
    KLMS = "klms"
    SOLVER = "solver"
    K_EULER = "k_euler"
    K_EULER_A = "k_euler_a"
    K_HEUN = "k_heun"


class TomeInfoModel(BaseModel):
    enable: bool = Field(False, description="Whether enable tomesd.")
    ratio: float = Field(0.5, description="The ratio of tokens to merge.")
    max_downsample: int = Field(
        1,
        description="Apply ToMe to layers with at most this amount of downsampling.",
    )
    sx: int = Field(2, description="The stride for computing dst sets.")
    sy: int = Field(2, description="The stride for computing dst sets.")
    seed: int = Field(
        -1,
        ge=-1,
        lt=2**32,
        description="""
Seed of the generation.
> If `-1`, then seed from `DiffusionModel` will be used.
> If `DiffusionModel.seed` is also `-1`, then random seed will be used.
""",
    )
    use_rand: bool = Field(True, description="Whether allow random perturbations.")
    merge_attn: bool = Field(True, description="Whether merge attention.")
    merge_crossattn: bool = Field(False, description="Whether merge cross attention.")
    merge_mlp: bool = Field(False, description="Whether merge mlp.")


class DiffusionModel(CallbackModel):
    use_circular: bool = Field(
        False,
        description="Whether should we use circular pattern (e.g. generate textures).",
    )
    seed: int = Field(
        -1,
        ge=-1,
        lt=2**32,
        description="""
Seed of the generation.
> If `-1`, then random seed will be used.
""",
    )
    variation_seed: int = Field(
        0,
        ge=0,
        lt=2**32,
        description="""
Seed of the variation generation.
> Only take effects when `variation_strength` is larger than 0.
""",
    )
    variation_strength: float = Field(
        0.0,
        ge=0.0,
        le=1.0,
        description="Strength of the variation generation.",
    )
    variations: List[VariationModel] = Field(
        default_factory=lambda: [],
        description="Variation ingredients",
    )
    num_steps: int = Field(20, description="Number of sampling steps", ge=5, le=100)
    guidance_scale: float = Field(
        7.5,
        description="Guidance scale for classifier-free guidance.",
    )
    negative_prompt: str = Field(
        "",
        description="Negative prompt for classifier-free guidance.",
    )
    is_anime: bool = Field(
        False,
        description="Whether should we generate anime images or not.",
    )
    version: MergedVersions = Field(
        MergedVersions.v1_5,
        description="Version of the diffusion model",
    )
    sampler: SDSamplers = Field(
        SDSamplers.K_EULER,
        description="Sampler of the diffusion model",
    )
    clip_skip: int = Field(
        -1,
        ge=-1,
        le=8,
        description="""
Number of CLIP layers that we want to skip.
> If it is set to `-1`, then `clip_skip` = 1 if `is_anime` else 0.
""",
    )
    custom_embeddings: Dict[str, List[List[float]]] = Field(
        {},
        description="Custom embeddings, often used in textual inversion.",
    )
    tome_info: TomeInfoModel = Field(TomeInfoModel(), description="tomesd settings.")


class CommonSDInpaintingModel(BaseModel):
    keep_original: bool = Field(
        False,
        description="Whether strictly keep the original image identical in the output image.",
    )
    use_raw_inpainting: bool = Field(
        False,
        description="""
Whether use the raw inpainting method.
> This is useful when you want to apply inpainting with custom SD models.
""",
    )
    raw_inpainting_fidelity: float = Field(
        0.2,
        ge=0.0,
        le=1.0,
        description="The fidelity of the input image when using raw inpainting.",
    )
    ref_url: str = Field(
        "",
        description="""
The `cdn` / `cos` url of the reference image.
> `cos` url from cloud is preferred.
> If empty string is provided, we will not use the reference feature.  
""",
    )
    ref_fidelity: float = Field(
        0.2,
        description="Fidelity of the reference image (if provided)",
    )


class Txt2ImgModel(DiffusionModel, MaxWHModel, TextModel):
    pass


class Img2ImgModel(MaxWHModel, ImageModel):
    pass


class Img2ImgDiffusionModel(DiffusionModel, Img2ImgModel):
    pass


class ControlStrengthModel(BaseModel):
    control_strength: float = Field(1.0, description="The strength of the control.")


class ReturnArraysModel(BaseModel):
    return_arrays: bool = Field(
        False,
        description="Whether return List[np.ndarray] directly, only for internal usages.",
    )


class _ControlNetModel(BaseModel):
    hint_url: str = Field(
        "",
        description="""
The `cdn` / `cos` url of the user's hint image.
> If empty string is provided, we will use `url` as `hint_url`.
> `cos` url from `qcloud` is preferred.
""",
    )
    hint_starts: Dict[str, float] = Field(
        default_factory=lambda: {},
        description="start ratio of each hint",
    )
    prompt: str = Field(..., description="Prompt.")
    fidelity: float = Field(
        0.05,
        ge=0.0,
        le=1.0,
        description="The fidelity of the input image, only take effects when `use_img2img` is True.",
    )
    use_img2img: bool = Field(True, description="Whether use img2img method.")
    num_samples: int = Field(1, ge=1, le=4, description="Number of samples.")
    bypass_annotator: bool = Field(False, description="Bypass the annotator.")
    base_model: MergedVersions = Field(
        MergedVersions.v1_5,
        description="The base model.",
    )
    guess_mode: bool = Field(False, description="Guess mode.")
    use_audit: bool = Field(False, description="Whether audit the outputs.")


class ControlNetModel(
    ReturnArraysModel,
    DiffusionModel,
    MaxWHModel,
    _ControlNetModel,
    ImageModel,
):
    pass


def handle_diffusion_model(m: DiffusionAPI, data: DiffusionModel) -> Dict[str, Any]:
    seed = None if data.seed == -1 else data.seed
    variation_seed = None
    variation_strength = None
    if data.variation_strength > 0:
        variation_seed = data.variation_seed
        variation_strength = data.variation_strength
    if data.variations is None:
        variations = None
    else:
        variations = [(v.seed, v.strength) for v in data.variations]
    m.switch_circular(data.use_circular)
    unconditional_cond = [data.negative_prompt] if data.negative_prompt else None
    clip_skip = data.clip_skip
    if clip_skip == -1:
        if data.is_anime or data.version.startswith("anime"):
            clip_skip = 1
        else:
            clip_skip = 0
    tome_info = data.tome_info.dict()
    enable_tome = tome_info.pop("enable")
    if not enable_tome:
        m.set_tome_info(None)
    else:
        if tome_info["seed"] == -1:
            if seed is None:
                tome_info.pop("seed")
            else:
                tome_info["seed"] = seed
        m.set_tome_info(tome_info)
    return dict(
        seed=seed,
        variation_seed=variation_seed,
        variation_strength=variation_strength,
        variations=variations,
        num_steps=data.num_steps,
        unconditional_guidance_scale=data.guidance_scale,
        unconditional_cond=unconditional_cond,
        sampler=data.sampler,
        verbose=verbose(),
        clip_skip=clip_skip,
        custom_embeddings=data.custom_embeddings or None,
    )


class GetPromptModel(BaseModel):
    text: str
    need_translate: bool = Field(
        True,
        description="Whether we need to translate the input text.",
    )


class GetPromptResponse(BaseModel):
    text: str
    success: bool
    reason: str


def endpoint2algorithm(endpoint: str) -> str:
    return endpoint[1:].replace("/", ".")


class IAlgorithm(AlgorithmBase, metaclass=ABCMeta):
    model_class: Type[BaseModel]
    response_model_class: Optional[Type[BaseModel]] = None
    last_latencies: Dict[str, float] = {}

    @classmethod
    def auto_register(cls) -> Callable[[AlgorithmBase], AlgorithmBase]:
        def _register(cls_: AlgorithmBase) -> AlgorithmBase:
            return cls.register(endpoint2algorithm(cls_.endpoint))(cls_)

        return _register

    def log_times(self, latencies: Dict[str, float]) -> None:
        super().log_times(latencies)
        self.last_latencies = latencies

    async def download_with_retry(self, url: str) -> bytes:
        return await download_with_retry(self.http_client.session, url)

    async def download_image_with_retry(self, url: str) -> Image.Image:
        return await download_image_with_retry(self.http_client.session, url)

    async def handle_diffusion_inpainting_model(
        self,
        data: CommonSDInpaintingModel,
    ) -> Dict[str, Any]:
        if not data.ref_url:
            reference = None
        else:
            reference = await self.download_image_with_retry(data.ref_url)
        return dict(
            use_raw_inpainting=data.use_raw_inpainting,
            raw_inpainting_fidelity=data.raw_inpainting_fidelity,
            reference=reference,
            reference_fidelity=data.ref_fidelity,
        )


# kafka


class Status(str, Enum):
    PENDING = "pending"
    WORKING = "working"
    FINISHED = "finished"
    EXCEPTION = "exception"
    INTERRUPTED = "interrupted"
    NOT_FOUND = "not_found"


# shortcuts


class SDParameters(BaseModel):
    is_anime: bool
    version: MergedVersions


def get_sd_from(data: SDParameters) -> ControlledDiffusionAPI:
    if not data.is_anime:
        version = data.version
    else:
        version = data.version if data.version.startswith("anime") else "anime"
    sd: ControlledDiffusionAPI = api_pool.get(APIs.SD)
    sd.switch_sd(version)
    sd.disable_control()
    return sd


__all__ = [
    "endpoint2algorithm",
    "DiffusionModel",
    "Txt2ImgModel",
    "Img2ImgModel",
    "Img2ImgDiffusionModel",
    "ReturnArraysModel",
    "ControlNetModel",
    "GetPromptModel",
    "GetPromptResponse",
    "Status",
    "IAlgorithm",
]
